theory CPO = Infra:

(*****************************************************************

         1. Complete partial order (cpo)
         2. Continuity
         3. Tarski Theorem

 *****************************************************************)

consts
  directed ::  "'a::order set => bool"

defs
  directed_def : "directed X == ((X ~= {}) &
                                 (ALL x y. x : X & y : X -->
                                    (EX z. z : X & x <= z & y <= z)))"

axclass bot0 < order
consts Bot :: "'a::bot0"  (* Bottom *)

axclass
  bot < bot0
  bottom_bot : "Bot <= (x::'a::bot0)"

axclass
  cpo < order
  complete_cpo : "(directed (X::'a::order set)) ==> (EX x::'a::order. X <<| x)"

(*** pointed cpo ***)

axclass cpo_bot < cpo, bot

consts  
  cont  :: "('a::cpo => 'b::cpo) => bool"
  adm   :: "('a::cpo => bool) => bool"

defs
  cont_def : "cont f == ALL X. 
                directed X --> f ` X <<| f (LUB X)"

  adm_def  : "adm P == ALL X. 
                directed X --> (ALL x:X. P x) --> P(LUB X)"

(********************************************************** 
                    small lemmas
 **********************************************************)

lemma complete_cpo_lm:
  "ALL (X::'a::cpo set). directed X --> (EX x. X <<| x)"
apply (simp add: complete_cpo)
done

(*** another definition of cont ***)

lemma cont_imp: "ALL f. cont f --> (ALL X. 
          directed X --> (EX x. ((f ` X) <<| (f x) & X <<| x)))"
apply (intro allI impI)
apply (insert complete_cpo_lm)
apply (drule_tac x="X" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="x" in exI)

apply (simp add: cont_def)
apply (drule_tac x="X" in spec)
apply (simp)

apply (insert LUB_lub_imp)
apply (drule_tac x="X" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

lemma cont_if: "ALL f. 
   (ALL X. directed X --> (EX x. ((f ` X) <<| (f x) & X <<| x))) --> cont f"
apply (intro allI impI)

apply (simp add: cont_def)
apply (intro allI impI)
apply (drule_tac x="X" in spec)
apply (simp)
apply (elim conjE exE)
apply (insert LUB_lub_imp)
apply (drule_tac x="X" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

(*** cont_iff ***)

lemma cont_iff: "cont f =
   (ALL X. directed X --> (EX x. ((f ` X) <<| (f x) & X <<| x)))"
apply (rule iffI)
apply (simp add: cont_imp)
apply (simp add: cont_if)
done

(*** cont_iffI ***)

lemma cont_iffI: 
  "(ALL X x. (directed X & X <<| x) --> (f ` X) <<| (f x))
    ==> cont f"
apply (simp add: cont_iff)
apply (intro allI impI)
apply (drule_tac x="X" in spec)
apply (simp)
apply (insert complete_cpo_lm)
apply (drule_tac x="X" in spec)
apply (simp)
apply (erule exE)
apply (rule_tac x="x" in exI)
apply (drule_tac x="x" in spec)
apply (simp)
done

(*** lub on cont ***)

lemma lub_cont_fun:
  "ALL X x f. directed X & X <<| x & cont f --> (f ` X) <<| f x"
apply (intro allI impI)
apply (simp add: cont_iff)
apply (elim conjE)
apply (drule_tac x="X" in spec)
apply (simp)
apply (elim conjE exE)
apply (insert lub_unique)
apply (drule_tac x="X" in spec)
apply (drule_tac x="xa" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

(*** lub on cont (e) ***)

lemma lub_cont_funE:
  "[|  X <<| x; directed X; cont f;
   [| f ` X <<| f x |] ==> R |]
   ==> R"
apply (simp add: lub_cont_fun)
done

(********************************************************** 
             lemmas and theorems for Tarski
 **********************************************************)

(* LUB is not affected by Bot *)

lemma lub_with_Bot: "ALL (x::('a::cpo_bot)) b f. 
  ({x. f x | x = Bot} <<| x) = ({x. f x} <<| x)"
apply (intro allI)
apply (rule iffI)

apply (simp add: is_lub_def is_ub_def)
apply (erule conjE)
apply (intro allI)
apply (intro impI)
apply (rotate_tac 1)
apply (drule_tac x="y" in spec)
apply (rotate_tac -1)
apply (drule mp)
apply (intro allI)
apply (drule_tac x="ya" in spec)
apply (simp add: bottom_bot)
apply (assumption)

apply (simp add: is_lub_def is_ub_def bottom_bot)
done

(************************************************************)

(* if A<=B then {x,y} is a directed set. *)
(* this is used in a proof for cont_mono *)

lemma directed_x_y: "ALL x y. x <= y --> directed {x, y}"
apply (intro allI,rule impI)
apply (unfold directed_def)
apply (simp)
apply (intro allI,rule impI)
apply (rule_tac x="y" in exI)
apply (auto)
done

(* the least upper bound of a set {x,y} *)
(* this is used in a proof for cont_mono *)

lemma lub_x_y: "ALL x y z. (x <= y & {x, y} <<| z ) --> z=y"
apply (intro allI impI)
apply (simp only: is_ub_def is_lub_def)
apply (elim conjE)
apply (drule_tac x="y" in spec, simp)
apply (drule_tac x="y" in spec, simp)
done

(************************ continuity *****************************)

(* f is continuous --> f is monotonic *)

lemma cont_mono: "ALL f. cont f --> mono f"
apply (intro allI, rule impI)
apply (simp add: cont_iff mono_def)
apply (intro allI impI)
apply (drule_tac x="{A,B}" in spec)
apply (simp add: directed_x_y)
apply (erule exE)
apply (erule conjE)
apply (insert lub_x_y)
apply (drule_tac x="A" in spec)
apply (drule_tac x="B" in spec)
apply (drule_tac x="x" in spec)
apply (simp add: is_lub_def is_ub_def)
done

(* directed and functions *)

lemma directed_cont: "ALL f X. cont f --> directed X --> directed (f ` X)"
apply (intro allI impI)
apply (insert cont_mono)
apply (drule_tac x="f" in spec)
apply (simp add: cont_iff directed_def)
apply (drule_tac x="X" in spec)
apply (simp)
apply (intro allI impI)
apply (elim conjE exE)
apply (insert in_image_set)
apply (drule_tac x="x" in spec)
apply (drule_tac x="f" in spec)
apply (drule_tac x="X" in spec)
apply (insert in_image_set)
apply (drule_tac x="y" in spec)
apply (drule_tac x="f" in spec)
apply (drule_tac x="X" in spec)
apply (simp)
apply (elim conjE exE)

apply (drule_tac x="ya" in spec)
apply (drule_tac x="yb" in spec)
apply (simp)
apply (elim conjE exE)

apply (rule_tac x="f z" in exI)
apply (simp)

apply (simp add: mono_def)
done

(* composition of continuous functions *)

lemma compo_cont: "ALL f g. cont f & cont g --> cont (g o f)"
apply (intro allI impI)
apply (erule conjE)
apply (insert directed_cont)
apply (drule_tac x="f" in spec)

apply (simp add: cont_iff)
apply (intro allI impI)
apply (drule_tac x="X" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="f ` X" in spec)
apply (simp)
apply (elim conjE exE)
apply (rule_tac x="x" in exI)
apply (simp)

apply (insert lub_unique)
apply (drule_tac x="f ` X" in spec)
apply (drule_tac x="xa" in spec)
apply (drule_tac x="f x" in spec)
apply (simp add: image_compose)
done

(*********************** Tarski continuous ****************************)

lemma Tarski_cont_directed_lm1:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)) n. 
     mono f --> (f ^ n) Bot <= (f ^ (n+1)) Bot"
apply (intro allI impI)
apply (induct_tac n)
apply (simp add: bottom_bot)
apply (simp add: mono_def)
done

lemma Tarski_cont_directed_lm2:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)) n m. 
     mono f --> (f ^ n) Bot <= (f ^ (n+m)) Bot"
apply (intro allI impI)
apply (induct_tac m)
apply (simp)
apply (insert Tarski_cont_directed_lm1)
apply (drule_tac x="f" in spec)
apply (drule_tac x="n+na" in spec)
apply (simp)
done

lemma Tarski_cont_directed:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)). 
     mono f --> directed {x. EX n. x = (f^n) Bot}"
apply (intro allI impI)
apply (simp add: directed_def)

apply (rule conjI)
apply (rule_tac x="Bot" in exI)
apply (rule_tac x="0" in exI)
apply (simp)

apply (intro allI impI)
apply (erule conjE)
apply (erule exE)
apply (erule exE)

apply (rename_tac f x y n m)
apply (case_tac "m <= n")

apply (rule_tac x="x" in exI)
apply (rule conjI)
apply (rule_tac x="n" in exI)
apply (assumption)
apply (simp)
apply (insert Tarski_cont_directed_lm2)
apply (drule_tac x="f" in spec)
apply (drule_tac x="m" in spec)
apply (drule_tac x="n-m" in spec)
apply (simp)

apply (rule_tac x="y" in exI)
apply (rule conjI)
apply (rule_tac x="m" in exI)
apply (assumption)
apply (drule_tac x="f" in spec)
apply (drule_tac x="n" in spec)
apply (drule_tac x="m-n" in spec)
apply (simp)
done

lemma Tarski_cont_image_expand_lm:
  "f ` {x. EX n. x = (f ^ n) Bot} = {x. EX n. x = f ((f ^ n) Bot)}"
apply (auto simp add: image_Collect)
done

lemma Tarski_cont_by_continuity:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)). cont f --> 
    (EX x. {x. EX n. x = f((f^n) Bot)} <<| f x &
           {x. EX n. x =   (f^n) Bot } <<| x)"
apply (intro allI impI)
apply (insert cont_mono)
apply (drule_tac x="f" in spec)
apply (simp)

apply (simp add: cont_iff)
apply (drule_tac x="{x. EX n. x = (f ^ n) Bot}" in spec)
apply (simp add: Tarski_cont_directed)
apply (erule exE)
apply (rule_tac x="x" in exI)
apply (simp add: Tarski_cont_image_expand_lm)
done

(******)

lemma Tarski_cont_lub_lm:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)) x.
     (EX n. x = (f ^ n) Bot) = ((EX n. x = (f ^ (n + 1)) Bot) | x = Bot)"
apply (intro allI)
apply (rule iffI)

(* ==> *)
apply (erule exE)
apply (insert nat_pos_zero)
apply (drule_tac x="n" in spec)
apply (erule disjE)
apply (simp)

apply (erule exE)
apply (rule disjI1)
apply (rule_tac x="m" in exI)
apply (simp)

(* <== *)
apply (erule disjE)
apply (erule exE)
apply (rule_tac x="n+1" in exI)
apply (simp)
apply (rule_tac x="0" in exI)
apply (simp)
done

(******)

lemma Tarski_cont_lub:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)) x.
    {x. EX n. x = f((f^n) Bot)} <<| x =
    {x. EX n. x =   (f^n) Bot } <<| x"
apply (simp add: Tarski_cont_lub_lm lub_with_Bot)
done

(******)

lemma Tarski_cont_least_lm:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)).
     mono f --> (ALL y. y = f y --> (ALL n. (f^n) Bot <= y))"
apply (intro allI impI)
apply (induct_tac n)
apply (simp add: bottom_bot)
apply (simp add: mono_def)
apply (drule_tac x="(f ^ na) Bot" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
done

lemma Tarski_cont_least:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)) x.
     mono f --> {x. EX n. x = (f^n) Bot } <<| x
       --> (ALL y. y = f y --> x <= y)"
apply (intro allI impI)
apply (simp add: is_lub_def)
apply (erule conjE)
apply (drule_tac x="y" in spec)
apply (drule mp)
apply (simp add: is_ub_def)
apply (intro allI impI)
apply (erule exE)
apply (simp add: Tarski_cont_least_lm)
apply (assumption)
done

(******)

theorem Tarski_cont:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)). cont f -->
    (EX x. {x. EX n. x = (f^n) Bot } <<| x & x isLFP f)"
apply (intro allI impI)
apply (insert Tarski_cont_by_continuity)
apply (drule_tac x="f" in spec)
apply (simp add: Tarski_cont_lub)
apply (erule exE)
apply (rule_tac x="x" in exI)
apply (insert lub_unique)
apply (drule_tac x="{x. EX n. x = (f ^ n) Bot}" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="f x" in spec)
apply (simp add: is_lfp_def)
apply (insert cont_mono)
apply (drule_tac x="f" in spec)
apply (simp add: Tarski_cont_least)
done

theorem Tarski_cont_LUB:
  "ALL (f::('a::cpo_bot => 'a::cpo_bot)). cont f
     --> LUB {x. EX n. x = (f^n) Bot } isLFP f"
apply (intro allI impI)
apply (insert Tarski_cont)
apply (drule_tac x="f" in spec)
apply (auto)
apply (insert LUB_lub_imp)
apply (drule_tac x="{x. EX n. x = (f^n) Bot }" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done


(********************************************************** 
           Fixed Point Induction (pointed CPO)
 **********************************************************)

theorem cpo_fixpoint_induction:
  "ALL (R::('a::cpo_bot => bool)) f. 
      cont f & adm R & R Bot & inductivefun R f 
        --> (EX y. (R y & y isLFP f))"
apply (auto)
apply (rule_tac x="LUB {x. EX n. x = (f^n) Bot }" in exI)
apply (simp add: Tarski_cont_LUB)
apply (simp add: adm_def)
apply (drule_tac x="{x. EX n. x = (f^n) Bot }" in spec)
apply (simp add: Tarski_cont_directed cont_mono)
apply (auto simp add: inductivefun_all_n)
done

(*** LFP version ***)

theorem cpo_fixpoint_induction_LFP:
  "ALL (R::('a::cpo_bot => bool)) f. 
      cont f & adm R & R Bot & inductivefun R f 
        --> R (LFP f)"
apply (auto)
apply (insert cpo_fixpoint_induction)
apply (drule_tac x="R" in spec)
apply (drule_tac x="f" in spec)
apply (simp)
apply (elim conjE exE)
apply (simp add: the_lfp)
done

(*****************************************************************)

end
