           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_Int_choice = Csp_semantics + Csp_Ext_choice + Csp_STOP:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*****************************************************************

         1. Union Fs: dom_F
         2. INTCHsf SF : dom_SF
         3. 
         4.

 *****************************************************************)

(*********************************************************
               Lemmas (Ext_choicef in Dom_F)
 *********************************************************)

(*** Int_choicef in dom_F ***)

lemma Int_choicef_in :
  "ALL Fs. (ALL F:Fs. F:dom_F) --> Union Fs: dom_F"
apply (intro allI impI)

apply (simp add: dom_F_def)
apply (simp (no_asm) add: CF1_def)
apply (intro allI impI)

apply (erule conjE)
apply (erule bexE)
apply (rename_tac Fs s X Y F)

apply (drule_tac x="F" in bspec)
apply (simp)
apply (rule_tac x="F" in bexI)
apply (blast intro: CF1_propI)
apply (simp)
done

(*** Int_choicef in dom_T Rep ***)

lemma Int_choicef_in_rep : 
  "ALL Fs. Union (Rep_dom_F ` Fs): dom_F"
apply (intro allI)
apply (insert Int_choicef_in)
apply (drule_tac x="Rep_dom_F ` Fs" in spec)
apply (simp add: Rep_dom_F)
done

(*********************************************************
               Lemmas (Ext_choicesf in Dom_SF)
 *********************************************************)

(* T2 *)

lemma Int_choicesf_in_T2 :
  "ALL TFs. TFs ~= {} & (ALL TF:TFs. TF:dom_SF) --> CT2 (INTCHsf TFs)"
apply (intro allI impI)
apply (simp add: Int_choicesf_def CT2_def)
apply (intro allI impI)
apply (simp add: Ext_choicet_in Rep_dom_T Abs_dom_T_inverse 
                 pair_fst_set_nonemp)
apply (simp add: Int_choicef_in Rep_dom_F Abs_dom_F_inverse)

apply (erule conjE)+
apply (erule bexE)
apply (rename_tac TFs s TF)

apply (simp add: dom_SF_def CT2_def)
apply (drule_tac x="TF" in bspec)
apply (simp)

apply (erule conjE)+
apply (drule_tac x="s" in spec)

  (* Evset *)

apply (rule conjI)
apply (rule_tac x="TF" in bexI)
apply (simp)
apply (simp)

  (* EvsetTick *)

apply (rule_tac x="TF" in bexI)
apply (simp)
apply (simp)
done

(* F2 *)

lemma Int_choicesf_in_F2 :
  "ALL TFs. TFs ~= {} & (ALL TF:TFs. TF:dom_SF) --> CF2 (INTCHsf TFs)"
apply (intro allI impI)
apply (simp add: Int_choicesf_def CF2_def)
apply (intro allI impI)
apply (simp add: Ext_choicet_in Rep_dom_T Abs_dom_T_inverse 
                 pair_fst_set_nonemp)
apply (simp add: Int_choicef_in Rep_dom_F Abs_dom_F_inverse)

apply (erule conjE)+
apply (erule bexE)
apply (rename_tac TFs s X Y TF)

apply (simp add: dom_SF_def CF2_def)
apply (drule_tac x="TF" in bspec)
apply (simp)
apply (erule conjE)+
apply (drule_tac x="s" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)

apply (rule_tac x="TF" in bexI)
apply (simp)
apply (simp)
done

(* F3 *)

lemma Int_choicesf_in_F3 :
  "ALL TFs. TFs ~= {} & (ALL TF:TFs. TF:dom_SF) --> CF3 (INTCHsf TFs)"
apply (intro allI impI)
apply (simp add: Int_choicesf_def CF3_def)
apply (intro allI impI)
apply (simp add: Ext_choicet_in Rep_dom_T Abs_dom_T_inverse 
                 pair_fst_set_nonemp)
apply (simp add: Int_choicef_in Rep_dom_F Abs_dom_F_inverse)

apply (erule exE)
apply (erule bexE)
apply (rename_tac TFs s X TF)

apply (simp add: dom_SF_def CF3_def)

apply (erule conjE)+
apply (drule_tac x="TF" in bspec)
apply (simp)

apply (erule conjE)+
apply (drule_tac x="s" in spec)
apply (drule mp)
apply (rule_tac x="X" in exI)
apply (simp)

apply (rule_tac x="TF" in bexI)
apply (simp)
apply (simp)
done

(*** Int_choicesf in dom_SF ***)

lemma Int_choicesf_in :
  "ALL TFs. TFs ~= {} & (ALL TF:TFs. TF:dom_SF) --> (INTCHsf TFs): dom_SF"
apply (intro allI impI)
apply (simp (no_asm) add: dom_SF_def)
apply (simp add: Int_choicesf_in_T2)
apply (simp add: Int_choicesf_in_F2)
apply (simp add: Int_choicesf_in_F3)
done

(*********************************************************
                     Lemmas (mono)
 *********************************************************)

(*** SF ***)

lemma Int_choicesf_mono: 
  "ALL SEs TFs. 
     SEs ~= {} & TFs ~= {} &
     (ALL SE : SEs. EX TF : TFs. SE <= TF)
                --> INTCHsf SEs <= INTCHsf TFs"
apply (intro allI impI)
apply (simp add: Int_choicesf_def)
apply (simp add: order_pair_def)

      (* T : dom *)

apply (insert Ext_choicet_in)
apply (frule_tac x="Rep_dom_T ` fst ` SEs " in spec)
apply (drule_tac x="Rep_dom_T ` fst ` TFs " in spec)
apply (simp add: Rep_dom_T)

      (* F : dom *)

apply (insert Int_choicef_in)
apply (frule_tac x="Rep_dom_F ` snd ` SEs " in spec)
apply (drule_tac x="Rep_dom_F ` snd ` TFs " in spec)
apply (simp add: Rep_dom_F)

apply (simp add: subset_T_def subset_F_def)
apply (simp add: Abs_dom_T_inverse Abs_dom_F_inverse)

      (* T *)

apply (insert Union_mono_all_ex)
apply (drule_tac x="Rep_dom_T ` fst ` SEs" in spec)
apply (drule_tac x="Rep_dom_T ` fst ` TFs" in spec)
apply (drule mp)

apply (elim conjE)
apply (rule ballI)
apply (simp add: image_iff)
apply (erule bexE)
apply (drule_tac x="xa" in bspec)
apply (simp)
apply (erule bexE)

apply (rule_tac x="TF" in bexI)
apply (simp)
apply (simp)

      (* F (symmetic with T) *)

apply (insert Union_mono_all_ex)
apply (drule_tac x="Rep_dom_F ` snd ` SEs" in spec)
apply (drule_tac x="Rep_dom_F ` snd ` TFs" in spec)
apply (drule mp)

apply (elim conjE)
apply (rule ballI)
apply (simp add: image_iff)
apply (erule bexE)
apply (drule_tac x="xa" in bspec)
apply (simp)
apply (erule bexE)

apply (rule_tac x="TF" in bexI)
apply (simp)
apply (simp)

      (**)

apply (simp)
done

(*********************************************************
                  Int choice trace eval
 *********************************************************)

lemma Int_choicet_eval:
  "ALL P Q ev. [[P |~| Q]]T ev = [[P]]T ev Un [[Q]]T ev"
apply (intro allI)
apply (simp add: eval_T_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Int_choicesf_def)

apply (insert Ext_choicet_in_rep)
apply (drule_tac x="{Abs_dom_T ([[P]]T ev), Abs_dom_T ([[Q]]T ev)}" in spec)
apply (simp add: Abs_dom_T_inverse)
done

(*********************************************************
                  Int choice failure eval
 *********************************************************)

lemma Int_choicef_eval:
  "ALL P Q ev. [[P |~| Q]]F ev = [[P]]F ev Un [[Q]]F ev"
apply (intro allI)
apply (simp add: eval_F_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Int_choicesf_def)

apply (insert Int_choicef_in_rep)
apply (drule_tac x="{Abs_dom_F ([[P]]F ev), Abs_dom_F ([[Q]]F ev)}" in spec)
apply (simp add: Abs_dom_F_inverse)
done

(*********************************************************
                    Int bind trace eval
 *********************************************************)

lemma Int_bindt_eval:
  "ALL X Pf ev. [[! :X .. Pf]]T ev 
                 = insert []t (Union {[[Pf a]]T ev |a. a:X})"
apply (intro allI)
apply (simp add: eval_T_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Int_choicesf_def)
apply (simp add: DIVsf_def)
apply (simp add: image_Collect_fun)
apply (simp add: STOPt_in Abs_dom_T_inverse)
apply (simp add: STOPt_def)

apply (insert Ext_choicet_in)
apply (drule_tac x=
  "insert {[]t}
          {Rep_dom_T (Abs_dom_T ([[Pf a]]T ev)) |a. a : X}" in spec)
apply (simp)
apply (drule mp)
apply (simp add: nil_t_set_in)
apply (intro allI impI)
apply (elim conjE exE)
apply (simp add: Rep_dom_T)

apply (simp add: Abs_dom_T_inverse)
done

(*********************************************************
             Parameterized Int choice failure eval
 *********************************************************)

lemma Int_bindf_eval:
  "ALL X Pf ev. [[! :X .. Pf]]F ev = Union {[[Pf a]]F ev |a. a:X}"
apply (intro allI)
apply (simp add: eval_F_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Int_choicesf_def)
apply (simp add: DIVsf_def)
apply (simp add: image_Collect_fun)
apply (simp add: emptyset_in_dom_F Abs_dom_F_inverse)

apply (insert Int_choicef_in_rep)
apply (drule_tac x="{Abs_dom_F ([[Pf a]]F ev) |a. a : X}" in spec)
apply (simp add: image_Collect_fun)
apply (simp add: Abs_dom_F_inverse)
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
