           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_Prefix = Csp_semantics:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*****************************************************************

         1. a ->t T  : dom_T
         2. a ->f F  : dom_F
         3. a ->sf TF: dom_SF
         4.

 *****************************************************************)

(*********************************************************
               Lemmas (Prefixt in Dom_T)
 *********************************************************)

(*** Prefixt in dom_T ***)

lemma Prefixt_in : "ALL a T. T:dom_T --> a ->t T: dom_T"
apply (intro allI impI)

apply (simp add: Prefixt_def)

apply (simp add: dom_T_def CT1_def)
apply (simp (no_asm) add: prefix_closed_def)
apply (intro allI impI)

apply (erule exE)
apply (erule conjE)+

  (* t = [] *)

apply (erule disjE)
apply (simp add: prefix_of_nil_t)

  (* t = [Ev a]t @t s *)

apply (erule exE)
apply (erule conjE)+

apply (insert trace_nil_or_Tick_or_Ev)
apply (drule_tac x="s" in spec)

  (* s = [] *)

apply (erule disjE)
apply (simp)

  (* s = [Tick]t --> inconsistent *)

apply (erule disjE)
apply (simp add: prefix_iff)
apply (erule exE)
apply (simp add: rep_appt_to_apprep event_in_trace Abs_trace_inverse)
apply (simp add: event_app_in_trace Rep_trace Abs_trace_inverse)

  (* s = [Ev aa]t @t sb *)

apply (erule exE)+
apply (rule disjI2)
apply (rule_tac x="sb" in exI)

apply (simp add: prefix_iff)
apply (erule exE)

apply (simp add: rep_appt_to_apprep event_in_trace Abs_trace_inverse)
apply (simp add: event_app_in_trace Rep_trace Abs_trace_inverse)

    (* sa:T --> sb:T because prefix_closed T *)

apply (unfold prefix_closed_def)
apply (drule_tac x="sb" in spec)
apply (drule_tac x="sa" in spec)
apply (drule mp)
apply (simp add: prefix_list_def)
apply (insert ALL_Rep_trace)
apply (drule_tac x="sa" in spec)
apply (simp_all)
done

(*********************************************************
               Lemmas (Prefixf in Dom_F)
 *********************************************************)

(*** Prefixt in dom_F ***)

lemma Prefixf_in : "ALL a F. F:dom_F --> a ->f F: dom_F"
apply (intro allI impI)

apply (simp add: Prefixf_def)
apply (simp add: dom_F_def)

apply (simp (no_asm) add: CF1_def)
apply (intro allI impI)

  (* s = []t *)

apply (erule conjE)
apply (erule disjE)
apply (blast)

  (* s = [Ev a]t @t sa *)

apply (rule disjI2)
apply (erule exE)
apply (rule_tac x="sa" in exI)

apply (unfold CF1_def)
apply (drule_tac x="sa" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (simp)
done


(*********************************************************
               Lemmas (Prefixsf in Dom_SF)
 *********************************************************)

(* T2 *)

lemma Prefixsf_in_T2 : "ALL a TF. TF:dom_SF --> CT2 (a ->sf TF)"
apply (intro allI impI)
apply (simp add: Prefixsf_def CT2_def)
apply (intro allI impI)
apply (simp add: Prefixt_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Prefixf_in Rep_dom_F Abs_dom_F_inverse)

apply (simp add: Prefixt_def)
apply (erule conjE)+
apply (erule disjE)

  (* s @t [Tick]t = []t --> inconsistent *)

apply (simp add: app_t_nil)
apply (simp add: nil_in_trace event_in_trace Abs_trace_inject)

  (* s @t [Tick]t = [Ev a]t @t sa *)

apply (erule exE)
apply (insert trace_nil_or_Tick_or_Ev)
apply (drule_tac x="s" in spec)

    (* s = [] --> inconsistent *)

apply (erule disjE)
apply (simp add: event_in_trace nil_in_trace Abs_trace_inverse)
apply (simp add: event_in_trace event_app_in_trace Rep_trace Abs_trace_inject)

    (* s = [Tick] --> inconsistent (with notick s) *)

apply (erule disjE)
apply (simp add: event_in_trace Abs_trace_inverse)

    (* s = [Ev a]t @t sa *)

apply (erule exE)+
apply (insert trace_app_in_trace_if)
apply (drule_tac x="sb" in spec)
apply (drule_tac x="[Tick]t" in spec)

apply (simp add: Prefixf_def)

      (* [] part of F2 *)

apply (rule conjI)
apply (rule disjI2)
apply (rule_tac x="sb" in exI)

apply (simp add: dom_SF_def CT2_def)
apply (erule conjE)
apply (drule_tac x="sb" in spec)

             (* trace --> list *)

apply (simp add: rep_appt_to_apprep event_in_trace Abs_trace_inverse)
apply (simp add: event_app_in_trace Rep_trace Abs_trace_inverse)
apply (simp add: event_app_in_trace Rep_trace Abs_trace_inject)
apply (simp add: Rep_trace_inverse)

      (* [Tick] part of F2 *)

apply (rule disjI2)
apply (rule_tac x="sa" in exI)

apply (simp add: dom_SF_def CT2_def)
apply (erule conjE)
apply (drule_tac x="sb" in spec)

             (* trace --> list *)

apply (simp add: rep_appt_to_apprep event_in_trace Abs_trace_inverse)
apply (simp add: event_app_in_trace Rep_trace Abs_trace_inverse)
apply (simp add: event_app_in_trace Rep_trace Abs_trace_inject)
apply (simp add: Rep_trace_inverse)
done

(* F2 *)

lemma Prefixsf_in_F2 : "ALL a TF. TF:dom_SF --> CF2 (a ->sf TF)"
apply (intro allI impI)
apply (simp add: Prefixsf_def CF2_def)
apply (intro allI impI)
apply (simp add: Prefixt_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Prefixf_in Rep_dom_F Abs_dom_F_inverse)

apply (simp add: Prefixf_def)
apply (erule conjE)+

  (* s @t [Tick]t = []t *)

apply (erule disjE)
apply (simp)
apply (case_tac "Ev a : Y")

    (* Ev a : Y --> inconsistent *)
apply (drule_tac x="Ev a" in spec)
apply (simp add: Prefixt_def)
apply (erule conjE)+
apply (drule_tac x="[]t" in spec)

apply (simp add: event_in_trace nil_in_trace Abs_trace_inverse)
apply (simp add: nil_t_in_all_dom Rep_dom_T)

    (* therefore Ev a ~: Y *)

apply (simp)

  (* s = [Ev a]t @t sa *)

apply (erule exE)+
apply (rename_tac a TF s X Y t)

apply (case_tac "EX b. b : Y & t @t [b]t : Rep_dom_T (fst TF)")

  (* EX b. b : Y & t @t [b]t : Rep_dom_T (fst TF) --> inconsist *)
apply (simp)

apply (erule exE)
apply (drule_tac x="b" in spec)

apply (simp add: Prefixt_def)
apply (erule conjE)+
apply (drule_tac x="t @t [b]t" in spec)

apply (insert trace_app_in_trace_if)
apply (rotate_tac -1)
apply (drule_tac x="t" in spec)
apply (rotate_tac -1)
apply (drule_tac x="[b]t" in spec)

apply (simp add: event_in_trace nil_in_trace Abs_trace_inverse)
apply (simp add: event_in_trace event_app_in_trace Rep_trace 
                 Abs_trace_inverse Abs_trace_inject Rep_trace_inverse)

  (* EX b. b : Y & t @t [b]t : Rep_dom_T (fst TF) --> inconsist *)

apply (simp add: dom_SF_def CF2_def)
apply (erule conjE)+

apply (rotate_tac -2)
apply (drule_tac x="t" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Y" in spec)
apply (drule mp)

apply (simp add: event_in_trace event_app_in_trace 
                 Rep_trace Abs_trace_inverse)
apply (intro allI impI)

apply (rotate_tac 3)
apply (drule_tac x="aa" in spec)
apply (simp)

   (* *)

apply (rule disjI2)
apply (rule_tac x="t" in exI)
apply (simp)
done

(* F3 *)

lemma Prefixsf_in_F3 : "ALL a TF. TF:dom_SF --> CF3 (a ->sf TF)"
apply (intro allI impI)
apply (simp add: Prefixsf_def CF3_def)
apply (intro allI impI)
apply (simp add: Prefixt_in Rep_dom_T Abs_dom_T_inverse)
apply (simp add: Prefixf_in Rep_dom_F Abs_dom_F_inverse)

apply (simp add: Prefixf_def)
apply (erule exE)

  (* s = []t *)

apply (erule disjE)
apply (simp add: Prefixt_def)

  (* s = [Ev a]t @t sa *)

apply (erule exE)
apply (simp add: Prefixt_def)

apply (rule disjI2)
apply (rule_tac x="sa" in exI)

apply (simp add: dom_SF_def CF3_def)
apply (erule conjE)+

apply (drule_tac x="sa" in spec)
apply (drule mp)
apply (rule_tac x="X" in exI)
apply (simp_all)
done

(*** Prefixsf in dom_SF ***)

lemma Prefixsf_in : "ALL a TF. TF:dom_SF --> (a ->sf TF): dom_SF"
apply (intro allI impI)
apply (simp (no_asm) add: dom_SF_def)
apply (simp add: Prefixsf_in_T2)
apply (simp add: Prefixsf_in_F2)
apply (simp add: Prefixsf_in_F3)
done

(*********************************************************
                     Lemmas (mono)
 *********************************************************)

lemma Prefixt_mono: "ALL a S T. S <= T --> (a ->t S) <= (a ->t T)"
apply (auto simp add: Prefixt_def)
done

lemma Prefixf_mono: "ALL a E F. E <= F --> (a ->f E) <= (a ->f F)"
apply (auto simp add: Prefixf_def)
done

lemma Prefixsf_mono: "ALL a SE TF. SE <= TF --> (a ->sf SE) <= (a ->sf TF)"
apply (auto simp add: Prefixsf_def)
apply (simp add: order_pair_def)
apply (simp add: subset_T_def subset_F_def)
apply (simp add: Prefixt_in Prefixf_in Rep_dom_T Rep_dom_F
                 Abs_dom_T_inverse Abs_dom_F_inverse)
apply (simp add: Prefixt_mono Prefixf_mono)
done

(*********************************************************
               Prefix choice trace eval
 *********************************************************)

lemma Prefixt_eval: "ALL a P ev. [[a -> P]]T ev = a ->t [[P]]T ev"
apply (intro allI)
apply (simp add: eval_T_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Prefixsf_def)
apply (simp add: Prefixt_in Abs_dom_T_inverse Rep_dom_T)
done

(*********************************************************
               Prefix choice failure eval
 *********************************************************)

lemma Prefixf_eval: "ALL a P ev. [[a -> P]]F ev = a ->f [[P]]F ev"
apply (intro allI)
apply (simp add: eval_F_def)
apply (simp add: eval_TF_decompo)
apply (simp add: Prefixsf_def)
apply (simp add: Prefixf_in Abs_dom_F_inverse Rep_dom_F)
done

(****************** to add them again ******************)

declare disj_not1 [simp]

end
