           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                August 2004                |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_law_basic = Csp_proof_step1     + Csp_proof_step2  +
                       Csp_proof_dist1     + Csp_proof_dist2  +
                       Csp_proof_Int       + Csp_proof_SKIP   + 
                       Csp_proof_fp        + Csp_proof_STOP:

(*****************************************************************

         1. equal <--> refinement
         2. reflexivity
         3. Transitivity

 *****************************************************************)

(*----------------------------------------------------------*
 |                                                          |
 |   decompose equality to refinement (Csp-Prover rules)    |
 |                                                          |
 *----------------------------------------------------------*)

lemma csp_eq_ref_iff:
  "(R1 =F[af] R2) = (R1 <=F[af] R2 & R2 <=F[af] R1)"
apply (simp add: refsf_rec_def eqsf_rec_def)
apply (auto intro: order_antisym)
done

lemma csp_eq_ref:
  "R1 =F[af] R2 ==> R1 <=F[af] R2"
apply (simp add: csp_eq_ref_iff)
done

lemma csp_ref_eq:
  "[| R1 <=F[af] R2 ; R2 <=F[af] R1 |] ==> R1 =F[af] R2"
apply (simp add: csp_eq_ref_iff)
done

(*-------------------------------------------------------*
 |                                                       |
 |           reflexivity (Csp-Prover rules)              |
 |                                                       |
 *-------------------------------------------------------*)

lemma eq_reflex: "R =F[af] R"
apply (simp add: eqsf_rec_def)
done

lemma ref_reflex: "R <=F[af] R"
apply (simp add: refsf_rec_def)
done

lemmas csp_reflex = eq_reflex ref_reflex

(*-------------------------------------------------------*
 |                                                       |
 |            Symmetry (Csp-Prover rules)                |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_sym:
  "R1 =F[af] R2 ==> R2 =F[af] R1"
apply (simp add: eqsf_rec_def)
done

lemma csp_symE:
  "[| R1 =F[af] R2 ; R2 =F[af] R1 ==> Z |] ==> Z"
apply (insert csp_sym[of R1 _ R2])
apply (simp)
done

(*-------------------------------------------------------*
 |                                                       |
 |           Transitivity (Csp-Prover rules)             |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_trans_eq: "[| R1 =F[af] R2 ; R2 =F[af] R3 |] ==> R1 =F[af] R3"
apply (simp add: eqsf_rec_def)
done

lemma csp_trans_ref: "[| R1 <=F[af] R2 ; R2 <=F[af] R3 |] ==> R1 <=F[af] R3"
apply (simp add: refsf_rec_def)
done

lemmas csp_trans = csp_trans_eq csp_trans_ref

(*-------------------------------------------------------*
 |                                                       |
 |              Rewrite (Csp-Prover rules)               |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_rw_left_eq:
  "[| R1 =F[af] R2 ; R2 =F R3 |] ==> R1 =F[af] R3"
apply (simp add: eqsf_rec_def)
done

lemma csp_rw_left_ref:
  "[| R1 =F[af] R2 ; R2 <=F R3 |] ==> R1 <=F[af] R3"
apply (simp add: eqsf_rec_def refsf_rec_def)
done

lemmas csp_rw_left = csp_rw_left_eq csp_rw_left_ref

lemma csp_rw_right_eq:
  "[| R3 =F[af] R2 ; R1 =F R2 |] ==> R1 =F[af] R3"
apply (simp add: eqsf_rec_def)
done

lemma csp_rw_right_ref:
  "[| R3 =F[af] R2 ; R1 <=F R2 |] ==> R1 <=F[af] R3"
apply (simp add: eqsf_rec_def refsf_rec_def)
done

lemmas csp_rw_right = csp_rw_right_eq csp_rw_right_ref

(*-------------------------------------------------------*
 |                                                       |
 |         Rewrite (<=F) (Csp-Prover rules)              |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_tr_left_eq: "[| R1 =F[af] R2 ; R2 =F R3 |] ==> R1 =F[af] R3"
apply (simp add: eqsf_rec_def)
done

lemma csp_tr_left_ref: "[| R1 <=F[af] R2 ; R2 <=F R3 |] ==> R1 <=F[af] R3"
apply (simp add: refsf_rec_def)
done

lemmas csp_tr_left = csp_tr_left_eq csp_tr_left_ref

lemma csp_tr_right_eq: "[| R2 =F[af] R3 ; R1 =F R2 |] ==> R1 =F[af] R3"
apply (simp add: eqsf_rec_def)
done

lemma csp_tr_right_ref: "[| R2 <=F[af] R3 ; R1 <=F R2 |] ==> R1 <=F[af] R3"
apply (simp add: refsf_rec_def)
done

lemmas csp_tr_right = csp_tr_right_eq csp_tr_right_ref


(*-------------------------------------------------------* 
 |                                                       |
 |       use these rules "csp_rule_..." as follows:      |
 |                                                       |
 |     lemmas new_rule = rule[THEN csp_rule_... ] or     |
 |     lemmas new_rule = csp_rule_...[OF rule]           |
 |                                                       |
 *-------------------------------------------------------*)

(*** make rw rules ***)

lemma csp_rule_rw_left_eq:
  "R1 =F[af1] R2 ==> (R1 =F[af2] R0) = (R2 =F[af2] R0)"
apply (simp add: eqsf_rec_def)
done

lemma csp_rule_rw_left_ref:
  "R1 =F[af1] R2 ==> (R1 <=F[af2] R0) = (R2 <=F[af2] R0)"
apply (simp add: eqsf_rec_def refsf_rec_def)
done

lemma csp_rule_rw_right_eq:
  "R1 =F[af1] R2 ==> (R0 =F[af2] R1) = (R0 =F[af2] R2)"
apply (simp add: eqsf_rec_def)
done

lemma csp_rule_rw_right_ref:
  "R1 =F[af1] R2 ==> (R0 <=F[af2] R1) = (R0 <=F[af2] R2)"
apply (simp add: eqsf_rec_def refsf_rec_def)
done

lemmas csp_rule_rw_left  = csp_rule_rw_left_eq csp_rule_rw_left_ref
lemmas csp_rule_rw_right = csp_rule_rw_right_eq csp_rule_rw_right_ref
lemmas csp_rule_rw       = csp_rule_rw_left csp_rule_rw_right

lemmas csp_rule_rev_rw   = csp_rule_rw[THEN sym]

(*** make tr rules ***)

lemma csp_rule_tr_left_eq:
  "R1 =F[af1] R2 ==> ((R2 =F[af2] R0) ==> (R1 =F[af2] R0))"
apply (simp add: eqsf_rec_def)
done

lemma csp_rule_tr_left_ref:
  "R1 <=F[af1] R2 ==> ((R2 <=F[af2] R0) ==> (R1 <=F[af2] R0))"
apply (simp add: eqsf_rec_def refsf_rec_def)
done

lemma csp_rule_tr_right_eq:
  "R2 =F[af1] R1 ==> ((R0 =F[af2] R2) ==> (R0 =F[af2] R1))"
apply (simp add: eqsf_rec_def)
done

lemma csp_rule_tr_right_ref:
  "R2 <=F[af1] R1 ==> ((R0 <=F[af2] R2) ==> (R0 <=F[af2] R1))"
apply (simp add: eqsf_rec_def refsf_rec_def)
done

lemmas csp_rule_tr_left  = csp_rule_tr_left_eq csp_rule_tr_left_ref
lemmas csp_rule_tr_right = csp_rule_tr_right_eq csp_rule_tr_right_ref
lemmas csp_rule_tr       = csp_rule_tr_left csp_rule_tr_right

lemmas csp_rule_tr_eq   = csp_rule_tr_left_eq csp_rule_tr_right_eq
lemmas csp_rule_tr_ref  = csp_rule_tr_left_ref csp_rule_tr_right_ref

(*** make intro rules (if active) ***)

lemma csp_rule_af_left_eq:
  "R1 =F:active R2 ==> ((R2 =F R0) ==> (R1 =F:active R0))"
apply (simp add: eqsf_rec_def)
done

lemma csp_rule_af_left_ref:
  "R1 <=F:active R2 ==> ((R2 <=F R0) ==> (R1 <=F:active R0))"
apply (simp add: refsf_rec_def)
done

lemma csp_rule_af_right_eq:
  "R2 =F:active R1 ==> ((R0 =F R2) ==> (R0 =F:active R1))"
apply (simp add: eqsf_rec_def)
done

lemma csp_rule_af_right_ref:
  "R2 <=F:active R1 ==> ((R0 <=F R2) ==> (R0 <=F:active R1))"
apply (simp add: refsf_rec_def)
done

lemmas csp_rule_af_left  = csp_rule_af_left_eq csp_rule_af_left_ref
lemmas csp_rule_af_right = csp_rule_af_right_eq csp_rule_af_right_ref
lemmas csp_rule_af       = csp_rule_af_left csp_rule_af_right

(*-------------------------------------------------------*
 |                                                       |
 |          Commutativity (Csp-Prover rules)             |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_Ext_choice_commut:
  "LET df IN P [+] Q =F[af] LET df IN Q [+] P"
apply (simp add: eqsf_iff eq_Ext_choice_commut)
done

lemma csp_Int_choice_commut:
  "LET df IN P |~| Q =F[af] LET df IN Q |~| P"
apply (simp add: eqsf_iff eq_Int_choice_commut)
done

lemma csp_Parallel_commut:
  "LET df IN P |[X]| Q =F[af] LET df IN Q |[X]| P"
apply (simp add: eqsf_iff eq_Parallel_commut)
done

lemmas csp_commut = csp_Ext_choice_commut csp_Int_choice_commut 
                    csp_Parallel_commut

(*----------------------------------------------------------*
 |                                                          |
 |              active flag (Csp-Prover rules)              |
 |                                                          |
 *----------------------------------------------------------*)

lemma csp_activate_eq:
  "R1 =F:active R2 ==> R1 =F[af] R2"
apply (simp add: eqsf_rec_def)
done

lemma csp_activate_ref:
  "R1 <=F:active R2 ==> R1 <=F[af] R2"
apply (simp add: refsf_rec_def)
done

lemma csp_inactivate_eq:
  "R1 =F R2 ==> R1 =F[af] R2"
apply (simp add: eqsf_rec_def)
done

lemma csp_inactivate_ref:
  "R1 <=F R2 ==> R1 <=F[af] R2"
apply (simp add: refsf_rec_def)
done

lemmas csp_activate = csp_activate_eq csp_activate_ref
lemmas csp_inactivate = csp_inactivate_eq csp_inactivate_ref

(*** make assumptions for both inactive and active ***)

lemma csp_asm_af_eqE:
  "[| R1 =F[af] R2 ; [| R1 =F R2 ; R1 =F:active R2 |] ==> Z |] ==> Z"
apply (simp add: eqsf_rec_def)
done

lemma csp_asm_af_refE:
  "[| R1 <=F[af] R2 ; [| R1 <=F R2 ; R1 <=F:active R2 |] ==> Z |] ==> Z"
apply (simp add: refsf_rec_def)
done

lemmas csp_asm_afE = csp_asm_af_eqE csp_asm_af_refE

end
