           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                August 2004                |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_law_step = Csp_law_decompo:

(*****************************************************************

         1. step law for STOP
         2. step law for Prefix
         3. step law for Ext_choice
         4. step law for Parallel
         5. step law for Hide
         6. step law for Rename
         7. step law for Sequence

 *****************************************************************)

(*********************************************************
                  Step law for STOP
 *********************************************************)

lemma csp_STOP_step: "LET df IN STOP =F[af] LET df IN ? x:{} -> DIV"
apply (simp add: eqsf_iff eq_STOP_step)
done

(*********************************************************
                  Step law for Prefix
 *********************************************************)

lemma csp_Prefix_step: 
 "LET df IN a -> P =F[af] LET df IN ? x:{a} -> P"
apply (simp add: eqsf_iff eq_Prefix_step)
done

(*********************************************************
                Step law for Ext_choice
 *********************************************************)

lemma csp_Ext_choice_step:
  "LET df IN (? :X -> Pf) [+] (? :Y -> Qf) =F[af]
   LET df IN Ext_choice_step X Y Pf Qf"
apply (simp add: eqsf_iff eq_Ext_choice_step)
done

(*********************************************************
                Step law for Parallel
 *********************************************************)

lemma csp_Parallel_step: 
  "LET df IN (? :Y -> Pf) |[X]| (? :Z -> Qf) =F[af]
   LET df IN Parallel_step X Y Z Pf Qf"
apply (simp add: eqsf_iff eq_Parallel_step)
done

(*********************************************************
                  Step law for Hide
 *********************************************************)

lemma csp_Hide_step: 
  "LET df IN (? :Y -> Pf) -- X =F[af]
   LET df IN Hide_step X Y Pf"
apply (simp add: eqsf_iff eq_Hide_step)
done

(*********************************************************
                Step law for Rename
 *********************************************************)

lemma csp_Rename_step: 
  "LET df IN (? :X -> Pf) [[r]] =F[af]
   LET df IN Rename_step r X Pf"
apply (simp add: eqsf_iff eq_Rename_step)
done

(*********************************************************
                Step law for Sequence
 *********************************************************)

lemma csp_Sequence_step: 
  "LET df IN (? :X -> Pf) ;; Q =F[af]
   LET df IN Sequence_step X Pf Q"
apply (simp add: eqsf_iff eq_Sequence_step)
done

(*********************************************************
                    Step laws
 *********************************************************)

lemmas csp_step = csp_STOP_step       csp_Prefix_step 
                      csp_Ext_choice_step csp_Parallel_step
                      csp_Hide_step       csp_Rename_step
                      csp_Sequence_step

lemmas csp_step_rw_left_eq   = csp_step[THEN csp_rule_rw_left_eq]
lemmas csp_step_rw_left_ref  = csp_step[THEN csp_rule_rw_left_ref]
lemmas csp_step_rw_right_eq  = csp_step[THEN csp_rule_rw_right_eq]
lemmas csp_step_rw_right_ref = csp_step[THEN csp_rule_rw_right_ref]

lemmas csp_step_rw_left      = csp_step_rw_left_eq csp_step_rw_left_ref
lemmas csp_step_rw_right     = csp_step_rw_right_eq csp_step_rw_right_ref

lemmas csp_step_rw           = csp_step_rw_left csp_step_rw_right

(*********************************************************
          light Step laws for Prefix and STOP
 *********************************************************)

lemmas csp_light_step = csp_STOP_step  csp_Prefix_step 

lemmas csp_light_step_rw_left_eq   = csp_light_step[THEN csp_rule_rw_left_eq]
lemmas csp_light_step_rw_left_ref  = csp_light_step[THEN csp_rule_rw_left_ref]
lemmas csp_light_step_rw_right_eq  = csp_light_step[THEN csp_rule_rw_right_eq]
lemmas csp_light_step_rw_right_ref = csp_light_step[THEN csp_rule_rw_right_ref]

lemmas csp_light_step_rw_left      = csp_light_step_rw_left_eq csp_light_step_rw_left_ref
lemmas csp_light_step_rw_right     = csp_light_step_rw_right_eq csp_light_step_rw_right_ref

lemmas csp_light_step_rw           = csp_light_step_rw_left csp_light_step_rw_right

end
