           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_proc = Csp_SKIP + Csp_STOP + Csp_DIV + Csp_Prefix 
                + Csp_Ext_choice + Csp_Int_choice + Csp_Parallel 
                + Csp_Hide + Csp_IF + Csp_Rename + Csp_Sequence:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite UnionT and InterT.                 *)
(*                  Union (B ` A) = (UN x:A. B x)                      *)
(*                  Inter (B ` A) = (INT x:A. B x)                     *)

declare Union_image_eq [simp del]
declare Inter_image_eq [simp del]

(*****************************************************************

         1. [[ proc ]] : dom_SF
         2. [[ proc ]] : mono
         3. 
         4. Timeout eval

 *****************************************************************)

(*********************************************************
                 ? :X -> Pf in Dom_SF
 *********************************************************)

lemma Pre_choicesf_in:
  "ALL X Pf ev. (ALL a. [[Pf a]]TF ev : dom_SF)
     --> [[? :X -> Pf]]TF ev : dom_SF"
apply (intro allI impI)
apply (simp)
apply (insert Ext_choicesf_in)
apply (drule_tac x="insert STOPsf {a ->sf [[Pf a]]TF ev |a. a : X}" in spec)
apply (simp add: STOPsf_in)
apply (drule mp)

apply (intro allI impI)
apply (erule exE)
apply (drule_tac x="aa" in spec)
apply (simp add: Prefixsf_in)
apply (simp)
done

lemma Pre_choicesf_in_induct:
  "!!ev set fun.
       (!!x. [[fun x]]TF ev : dom_SF)
       ==> EXTCHsf insert STOPsf {a ->sf [[fun a]]TF ev |a. a : set}
           : dom_SF"
apply (insert Pre_choicesf_in)
apply (drule_tac x="set" in spec)
apply (drule_tac x="fun" in spec)
apply (drule_tac x="ev" in spec)
apply (drule mp)
apply (intro allI)
apply (simp)
apply (simp)
done

(*********************************************************
                  ! :X .. Pf in Dom_SF
 *********************************************************)

lemma Int_bindsf_in:
  "ALL X Pf ev. (ALL a. [[Pf a]]TF ev : dom_SF)
     --> [[! :X .. Pf]]TF ev : dom_SF"
apply (intro allI impI)
apply (simp)
apply (insert Int_choicesf_in)
apply (drule_tac x="insert DIVsf {[[Pf a]]TF ev |a. a : X}" in spec)
apply (simp add: DIVsf_in)
apply (drule mp)

apply (intro allI impI)
apply (erule exE)
apply (drule_tac x="aa" in spec)
apply (simp_all)
done

lemma Int_bindsf_in_induct:
  "!!(ev::('a => 'c dom_SF)) (set::'c set)
       fun::'c => ('a, 'c) proc.
       (!!x. [[fun x]]TF ev : dom_SF)
       ==> INTCHsf insert DIVsf {[[fun a]]TF ev |a. a : set} : dom_SF"
apply (insert Int_bindsf_in)
apply (drule_tac x="set" in spec)
apply (drule_tac x="fun" in spec)
apply (drule_tac x="ev" in spec)
apply (simp)
done

(*********************************************************
             !!!    [[P]]TF in Dom_SF    !!!
 *********************************************************)

lemma procsf_in: "ALL P ev. [[P]]TF ev : dom_SF"
apply (intro allI)
apply (induct_tac P)
apply (simp add: STOPsf_in)
apply (simp add: SKIPsf_in)
apply (simp add: DIVsf_in)
apply (simp add: Prefixsf_in)
apply (simp add: Pre_choicesf_in_induct)
apply (simp add: Ext_choicesf_in)
apply (simp add: Int_choicesf_in)
apply (simp add: Int_bindsf_in_induct)
apply (simp)
apply (simp add: Parallelsf_in)
apply (simp add: Hidesf_in)
apply (simp add: Renamesf_in)
apply (simp add: Sequencesf_in)
apply (simp add: Rep_dom_SF)
done

(*********************************************************
                       T2 F1
 *********************************************************)

lemma T2_F1_Evset:
     "ALL s P X ev. notick s & X <= Evset & s @t [Tick]t : [[P]]T ev
           --> (s, X) : [[P]]F ev"
apply (intro allI impI)
apply (insert procsf_in)
apply (drule_tac x="P" in spec)
apply (drule_tac x="ev" in spec)

apply (simp add: eval_TF_decompo)
apply (simp add: dom_SF_def)
apply (simp add: CT2_def)

apply (erule conjE)+
apply (drule_tac x="s" in spec)
apply (simp add: Abs_dom_T_inverse eval_T_in)
apply (simp add: Abs_dom_F_inverse eval_F_in)

      (* F1 *)

apply (insert eval_F_in)
apply (drule_tac x="P" in spec)
apply (drule_tac x="ev" in spec)
apply (simp add: dom_F_def)
apply (unfold CF1_def)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Evset" in spec)
apply (drule_tac x="X" in spec)
apply (simp)
done

lemma T2_F1_EvsetTick:
     "ALL s P X ev. notick s & s @t [Tick]t : [[P]]T ev
           --> (s @t [Tick]t, X) : [[P]]F ev"
apply (intro allI impI)
apply (insert procsf_in)
apply (drule_tac x="P" in spec)
apply (drule_tac x="ev" in spec)

apply (simp add: eval_TF_decompo)
apply (simp add: dom_SF_def)
apply (simp add: CT2_def)

apply (erule conjE)+
apply (drule_tac x="s" in spec)
apply (simp add: Abs_dom_T_inverse eval_T_in)
apply (simp add: Abs_dom_F_inverse eval_F_in)

      (* F1 *)

apply (insert eval_F_in)
apply (drule_tac x="P" in spec)
apply (drule_tac x="ev" in spec)
apply (simp add: dom_F_def)
apply (unfold CF1_def)
apply (drule_tac x="s @t [Tick]t" in spec)
apply (drule_tac x="EvsetTick" in spec)
apply (drule_tac x="X" in spec)
apply (simp add: EvsetTick_def)
done

(*********************************************************
                    T2 F1 for Tick
 *********************************************************)

lemma T2_F1_Tick:
     "ALL P X ev. X <= Evset & [Tick]t : [[P]]T ev --> ([]t, X) : [[P]]F ev"
apply (intro allI impI)
apply (insert T2_F1_Evset)
apply (drule_tac x="[]t" in spec)
apply (drule_tac x="P" in spec)
apply (drule_tac x="X" in spec)
apply (drule_tac x="ev" in spec)
apply (simp add: trace_nil_app notick_nil)
done

(*********************************************************
                An order over procs
 *********************************************************)

(*** SF --> T_F ***)

lemma order_sf_tf:
  "ALL P Q ev. ([[P]]SF ev <= [[Q]]SF ev) = ([[P]]TF ev <= [[Q]]TF ev)"
apply (intro allI)
apply (simp add: eval_SF_def)
apply (simp add: subset_SF_def)
apply (simp add: procsf_in Abs_dom_SF_inverse)
done

lemma order_sf_tf_ev:
  "ALL P ev ev'. ([[P]]SF ev <= [[P]]SF ev') 
                 = ([[P]]TF ev <= [[P]]TF ev')"
apply (intro allI)
apply (simp add: eval_SF_def)
apply (simp add: subset_SF_def)
apply (simp add: procsf_in Abs_dom_SF_inverse)
done

(*** SF --> T and F ***)

lemma order_sf_decompo: 
  "ALL P Q ev. ([[P]]SF ev <= [[Q]]SF ev)
             = ([[P]]T ev <= [[Q]]T ev & [[P]]F ev <= [[Q]]F ev)"
apply (intro allI)
apply (simp add: eval_T_def eval_F_def)
apply (simp add: subset_SF_def)
apply (simp add: order_pair_def)
apply (simp add: subset_T_def subset_F_def eval_SF_def)
apply (simp add: procsf_in Abs_dom_SF_inverse)
done

(*********************************************************
                 Pre_choicesf mono
 *********************************************************)

lemma Pre_choicesf_mono: 
  "ALL X Pf ev ev'. 
      (ALL a. [[Pf a]]TF ev <= [[Pf a]]TF ev')
          --> [[? :X -> Pf]]TF ev <= [[? :X -> Pf]]TF ev'"
apply (intro allI impI)
apply (insert Ext_choicesf_mono)
apply (drule_tac x="insert STOPsf {a ->sf [[Pf a]]TF ev |a. a : X}" in spec)
apply (drule_tac x="insert STOPsf {a ->sf [[Pf a]]TF ev' |a. a : X}" in spec)
apply (simp del: split_paired_All split_paired_Ex)

   (* note: split_paired_All: (ALL x. ?P x) = (ALL a b. ?P (a, b)) *)
   (*       split_paired_Ex : (EX x. ?P x) = (EX a b. ?P (a, b))   *)

apply (drule mp)
apply (rule conjI)

apply (intro allI impI)
apply (rule disjI2)
apply (elim conjE exE)
apply (rule_tac x="a ->sf [[Pf a]]TF ev'" in exI)
apply (simp add: Prefixsf_mono)
apply (rule_tac x="a" in exI)
apply (simp)

apply (intro allI impI)
apply (rule disjI2)
apply (elim conjE exE)
apply (rule_tac x="a ->sf [[Pf a]]TF ev" in exI)
apply (simp add: Prefixsf_mono)
apply (rule_tac x="a" in exI)
apply (simp)

apply (simp)
done

lemma Pre_choicesf_mono_induct:
    "!!ev ev' set fun.
       (!!x. ev <= ev' --> [[fun x]]TF ev <= [[fun x]]TF ev')
       ==> ev <= ev' -->
           EXTCHsf insert STOPsf {a ->sf [[fun a]]TF ev |a. a : set}
           <= EXTCHsf insert STOPsf {a ->sf [[fun a]]TF ev' |a. a : set}"
apply (insert Pre_choicesf_mono)
apply (drule_tac x="set" in spec)
apply (drule_tac x="fun" in spec)
apply (drule_tac x="ev" in spec)
apply (drule_tac x="ev'" in spec)
apply (auto)
done

(*********************************************************
                    ! :X .. Pf mono
 *********************************************************)

lemma Int_bindsf_mono:
  "ALL X Pf ev ev'. 
      (ALL a. [[Pf a]]TF ev <= [[Pf a]]TF ev')
          --> [[! :X .. Pf]]TF ev <= [[! :X .. Pf]]TF ev'"
apply (intro allI impI)
apply (insert Int_choicesf_mono)
apply (drule_tac x="insert DIVsf {[[Pf a]]TF ev |a. a : X}" in spec)
apply (drule_tac x="insert DIVsf {[[Pf a]]TF ev' |a. a : X}" in spec)
apply (simp del: split_paired_All split_paired_Ex)

apply (drule mp)
apply (intro allI impI)
apply (rule disjI2)
apply (elim conjE exE)
apply (rule_tac x="[[Pf a]]TF ev'" in exI)
apply (simp)
apply (rule_tac x="a" in exI)
apply (simp_all)
done

lemma Int_bindsf_mono_induct:
  "!!(ev::'a => 'c dom_SF) (ev'::'a => 'c dom_SF)
       (set::'c set) fun::'c => ('a, 'c) proc.
       (!!x::'c. ev <= ev' --> [[fun x]]TF ev <= [[fun x]]TF ev')
       ==> ev <= ev' -->
           INTCHsf insert DIVsf {[[fun a]]TF ev |a::'c. a : set}
           <= INTCHsf insert DIVsf {[[fun a]]TF ev' |a::'c. a : set}"
apply (intro allI impI)
apply (insert Int_bindsf_mono)
apply (drule_tac x="set" in spec)
apply (drule_tac x="fun" in spec)
apply (drule_tac x="ev" in spec)
apply (drule_tac x="ev'" in spec)
apply (auto)
done

(*********************************************************
               !!!   proc sf mono   !!!
 *********************************************************)

lemma procsf_mono: 
  "ALL ev ev' P. ev <= ev' --> [[P]]SF ev <= [[P]]SF ev'"
apply (intro allI)
apply (induct_tac P)
apply (simp add: eval_SF_def)
apply (simp add: eval_SF_def)
apply (simp add: eval_SF_def)
apply (simp add: order_sf_tf_ev Prefixsf_mono)
apply (simp add: order_sf_tf_ev Pre_choicesf_mono_induct)
apply (simp add: order_sf_tf_ev Ext_choicesf_mono)
apply (simp add: order_sf_tf_ev Int_choicesf_mono)
apply (simp add: order_sf_tf_ev Int_bindsf_mono_induct)
apply (simp add: order_sf_tf_ev)
apply (simp add: order_sf_tf_ev Parallelsf_mono)
apply (simp add: order_sf_tf_ev Hidesf_mono)
apply (simp add: order_sf_tf_ev Renamesf_mono)
apply (simp add: order_sf_tf_ev Sequencesf_mono)

apply (simp add: eval_SF_def Rep_dom_SF_inverse)
apply (simp add: order_prod_def)
done

(*** mono function ***)

lemma procsf_mono_fun: "ALL P. mono [[P]]SF"
apply (simp add: procsf_mono mono_def)
done

(*********************************************************
                     Timeout trace eval
 *********************************************************)

lemma Timeoutt_eval:
  "ALL P Q ev. [[P [> Q]]T ev = [[P]]T ev Un [[Q]]T ev"
apply (intro allI)
apply (simp add: Ext_choicet_eval)
apply (simp add: Int_choicet_eval)
apply (simp add: STOPt_eval)
apply (simp add: STOPt_def)
apply (auto simp add: eval_T_in nil_t_in_all_dom)
done

(*********************************************************
                     Timeout failure eval
 *********************************************************)

(*** sub ***)

lemma Timeoutf_eval_subset:
  "ALL P Q ev. [[P [> Q]]F ev 
            <= [[Q]]F ev Un
                     {(s,X). s ~= []t & (s,X) : [[P]]F ev} Un
                     {([]t,X) |X. X <= Evset & [Tick]t : [[P]]T ev}"
apply (intro allI)
apply (simp add: Ext_choicef_eval)
apply (simp add: Int_choicef_eval)
apply (simp add: subset_def Ball_def)

apply (intro allI)
apply (rename_tac P Q ev s X)

      (* s ~= []t *)

apply (rule conjI)
apply (intro impI)
apply (simp add: STOPf_eval)
apply (simp add: STOPf_def)
apply (blast)

      (* s = []t *)

apply (intro impI)
apply (simp add: Int_choicet_eval)
apply (simp add: STOPt_eval)
apply (simp add: STOPt_def)

apply (erule conjE)+

      (* [Tick]t : [[P]]T ev *)

apply (erule disjE)
apply (simp)

      (* [Tick]t = []t --> contradiction *)

apply (erule disjE)
apply (simp add: nil_in_trace event_in_trace Abs_trace_inject)

apply (simp add: T2_F1_Tick subset_def Ball_def)
done

(*** sup ***)

lemma Timeoutf_eval_supset:
  "ALL P Q ev. [[Q]]F ev Un
           {(s,X). s ~= []t & (s,X) : [[P]]F ev} Un
           {([]t,X) |X. X <= Evset & [Tick]t : [[P]]T ev}
           <= [[P [> Q]]F ev"
apply (intro allI)
apply (simp add: Ext_choicef_eval)
apply (simp add: Int_choicef_eval)
apply (simp add: subset_def Ball_def)

apply (intro allI)
apply (rename_tac P Q ev s X)

      (* (s, X) : [[Q]]F ev *)

apply (rule conjI)
apply (intro allI impI)
apply (simp add: STOPf_eval)
apply (simp add: STOPf_def)
apply (simp add: EvsetTick_def)

      (* s = []t & (ALL x. x : X --> x : Evset) & [Tick]t : [[P]]T ev *)

apply (simp add: Int_choicet_eval)
done

(*** Timeout failure eval ***)

lemma Timeoutf_eval:
  "ALL P Q ev. [[P [> Q]]F ev 
             = [[Q]]F ev Un
                 {(s,X). s ~= []t & (s,X) : [[P]]F ev} Un
                 {([]t,X) |X. X <= Evset & [Tick]t : [[P]]T ev}"
apply (intro allI)
apply (rule equalityI)
apply (simp add: Timeoutf_eval_subset)
apply (simp add: Timeoutf_eval_supset)
done

(****************** to add them again ******************)

declare Union_image_eq [simp]
declare Inter_image_eq [simp]

end
