           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_proc_rec = Csp_map_proc:

(*****************************************************************

         1. rec proc : mono & contraction
         2. rec proc : unique fixpoint
         3. 
         4. 

 *****************************************************************)


(*** infinite sequence converging to the fixed point ***)

consts
  DIVsf_prod :: "('n,'a) dom_SF_prod"

defs
  DIVsf_prod_def: "DIVsf_prod == (%C. Abs_dom_SF DIVsf)"

consts
  seqsf  :: "('n,'a) procDef => ('n,'a) dom_SF_prod infinite_seq"
  seqsfn :: "('n,'a) procDef => nat => ('n,'a) dom_SF_prod"

defs
  seqsf_def  : "seqsf  df   == (%n. seqsfn df n)"
  seqsfn_def : "seqsfn df n == ([[df]]DF ^ n) DIVsf_prod"

(*********************************************************
                    proc def mono  
 *********************************************************)

lemma mono_procDef: 
  "ALL ev ev' df. ev <= ev' --> [[df]]DF ev  <= [[df]]DF ev'"
apply (intro allI impI)
apply (simp add: eval_procDef_def)
apply (simp (no_asm) add: order_prod_def)
apply (simp add: procsf_mono)
done

lemma mono_procDef_fun: "ALL df. mono [[df]]DF"
apply (simp add: mono_procDef mono_def)
done

(*********************************************************
                 rec proc non-expanding   
 *********************************************************)

lemma non_expand_procDef:
  "ALL df. (ALL C. nohide (df C)) --> non_expanding [[df]]DF"
apply (intro allI impI)
apply (insert prod_non_expand)
apply (drule_tac x="[[df]]DF" in spec)
apply (simp add: eval_procDef_def)
apply (simp add: exchange_args_def)
apply (drule mp)

apply (intro allI)
apply (insert non_expand_proc)
apply (drule_tac x="df i" in spec)
apply (simp_all)
done

(*********************************************************
                 rec proc contraction (1/2)
 *********************************************************)

lemma contract_procDef:
  "ALL df. (ALL C. nohide (df C) & guard (df C))
       --> contraction_alpha [[df]]DF (1/2)"
apply (intro allI impI)
apply (insert prod_contraction_alpha)
apply (drule_tac x="[[df]]DF" in spec)
apply (drule_tac x="1/2" in spec)
apply (simp add: eval_procDef_def)
apply (simp add: exchange_args_def)
apply (drule mp)

apply (intro allI)
apply (insert contract_proc)
apply (drule_tac x="df i" in spec)
apply (simp_all)
done

(*********************************************************
               Banach Theorem for dom_SF
 *********************************************************)

lemma Banach_dom_SF:
  "ALL df. (ALL C. nohide (df C) & guard (df C))
        --> (EX SF. (seqsf df convergeTo SF 
                    & SF = [[df]]DF SF
                    & (ALL TF. TF = [[df]]DF TF --> SF = TF)))"
apply (intro allI impI)
apply (insert Banach)
apply (drule_tac x="[[df]]DF" in spec)
apply (drule_tac x="seqsf df" in spec)
apply (drule mp)
apply (insert contract_procDef)
apply (drule_tac x="df" in spec)
apply (simp add: contraction_alpha_contraction)

apply (intro allI)
apply (simp add: seqsf_def seqsfn_def)
apply (simp)
done

(*********************************************************
               Banach Theorem for dom_SF UFP
 *********************************************************)

lemma Banach_dom_SF_UFP:
  "ALL df. (ALL C. nohide (df C) & guard (df C))
              --> [[df]]DF hasUFP"
apply (intro allI impI)
apply (insert Banach_dom_SF)
apply (drule_tac x="df" in spec)
apply (simp add: is_ufp_def has_UFP_def)
apply (erule exE)
apply (rule_tac x="SF" in exI)
apply (simp)
done

(*********************************************************
               Banach Theorem for FPdom_SF fixpoint
 *********************************************************)

lemma Banach_FPdom_SF_fixpoint:
  "ALL df. (ALL C. nohide (df C) & guard (df C))
        --> [[df]]DF (UFP [[df]]DF) = UFP [[df]]DF"
apply (intro allI impI)
apply (unfold UFP_def)
apply (simp add: Banach_dom_SF_UFP)

apply (insert Banach_dom_SF)
apply (drule_tac x="df" in spec)
apply (simp)
apply (erule exE)

apply (insert the_ufp)
apply (drule_tac x="SF" in spec)
apply (drule_tac x="[[df]]DF" in spec)
apply (simp add: is_ufp_def UFP_def Banach_dom_SF_UFP)
done

(*********************************************************
               Banach Theorem for FPdom_SF Unique
 *********************************************************)

lemma Banach_FPdom_SF_unique:
  "ALL df. (ALL C. nohide (df C) & guard (df C))
       --> (ALL TF. TF = [[df]]DF TF --> TF = UFP [[df]]DF)"
apply (intro allI impI)
apply (unfold UFP_def)
apply (simp add: Banach_dom_SF_UFP)
apply (insert the_ufp)
apply (drule_tac x="TF" in spec)
apply (drule_tac x="[[df]]DF" in spec)
apply (drule mp)

apply (simp add: is_ufp_def)
apply (intro allI impI)
apply (insert Banach_dom_SF)
apply (drule_tac x="df" in spec)
apply (simp)
apply (erule exE)

apply (erule conjE)+
apply (frule_tac x="TF" in spec)
apply (drule_tac x="y" in spec)
apply (simp)
apply (simp add: UFP_def Banach_dom_SF_UFP)
done

(*********************************************************
                       seqsf
 *********************************************************)

lemma Suc_seqsfn: "ALL df n. seqsfn df (Suc n) = [[df]]DF (seqsfn df n)"
apply (simp add: seqsfn_def)
done

(*********************************************************
                  increasing seqsf
 *********************************************************)

lemma inc_seqsf: "ALL df n. seqsfn df n <= seqsfn df (Suc n)"
apply (intro allI)
apply (induct_tac n)

      (* base *)
apply (simp add: seqsfn_def)
apply (simp add: DIVsf_prod_def)
apply (simp add: order_prod_def)
apply (simp add: DIVsf_def STOPt_def)
apply (simp add: subset_SF_def)
apply (simp add: BOT_in_dom_SF Abs_dom_SF_inverse)
apply (simp add: BOT_is_bottom_dom_SF Rep_dom_SF)

      (* step *)
apply (simp add: seqsfn_def)
apply (simp add: mono_procDef)
done

(*********************************************************
                     mono seqsf
 *********************************************************)

lemma mono_seqsf: "ALL df df' n. 
    (ALL ev. [[df]]DF ev <= [[df']]DF ev) --> seqsfn df n <= seqsfn df' n"
apply (intro allI)
apply (induct_tac n)

      (* base *)
apply (simp add: seqsfn_def)

      (* step *)
apply (intro allI impI)
apply (simp add: Suc_seqsfn)
apply (drule_tac x="seqsfn df na" in spec)
apply (insert mono_procDef)
apply (drule_tac x="seqsfn df na" in spec)
apply (drule_tac x="seqsfn df' na" in spec)
apply (drule_tac x="df'" in spec)
apply (simp)
done

(*********************************************************
                 proc compo procDef
 *********************************************************)

lemma compo_proc: "ALL P f. [[Rewrite P By f]]SF = [[P]]SF o [[f]]DF"
apply (intro allI)
apply (induct_tac P)
apply (simp_all add: eval_SF_def expand_fun_eq
                     procsf_in Abs_dom_SF_inject)
      (* process name *)
apply (intro allI)
apply (simp_all add: eval_procDef_def eval_SF_def)
apply (simp_all add: Rep_dom_SF_inverse)
done

end
