           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Csp_proof_step1 = Csp_relation_def:

(*****************************************************************

         1. step laws
         2.
         3. 
         4. 

 *****************************************************************)

consts
  Ext_choice_step  :: 
      "'a set => 'a set => ('a => ('n,'a) proc) 
                        => ('a => ('n,'a) proc) => ('n,'a) proc"

  Parallel_step ::
      "'a set => 'a set => 'a set
          => ('a => ('n,'a) proc) => ('a => ('n,'a) proc)
          => ('n,'a) proc"

  Hide_step ::
      "'a set => 'a set => ('a => ('n,'a) proc) => ('n,'a) proc"

defs
  Ext_choice_step_def :
   "Ext_choice_step X Y Pf Qf ==
      ? x:(X Un Y) -> 
                (IF (x : X & x : Y) THEN Pf x |~| Qf x
                 ELSE IF (x : X) THEN Pf x
                 ELSE Qf x)"
defs
  Parallel_step_def :
   "Parallel_step X Y Z Pf Qf ==
      ? x:((X Int Y Int Z) Un (Y - X) Un (Z - X))
         -> IF (x : X) THEN (Pf x |[X]| Qf x)
               ELSE IF (x : Y & x : Z) THEN ((Pf x |[X]| ? x:Z -> Qf x)
                                     |~| (? x:Y -> Pf x |[X]| Qf x))
               ELSE IF (x : Y) THEN (Pf x |[X]| ? x:Z -> Qf x)
               ELSE (? x:Y -> Pf x |[X]| Qf x)"
defs
  Hide_step_def :
   "Hide_step X Y Pf ==
      IF (Y Int X = {}) THEN (? x:Y -> (Pf x -- X))
                         ELSE ((? x:(Y-X) -> (Pf x -- X)) 
                               [> (! x:(Y Int X) .. (Pf x -- X)))"

(*********************************************************
                    stop expansion
 *********************************************************)

lemma eq_STOP_step: "STOP =F<ev> ? x:{} -> DIV"
apply (simp add: eqsf_decompo)
apply (rule conjI)
apply (simp add: STOPt_eval STOPt_def)
apply (simp add: Pre_choicet_eval)

apply (simp add: STOPf_eval STOPf_def)
apply (simp add: Pre_choicef_eval)
apply (simp add: EvsetTick_def)
done

(*********************************************************
                    Prefix expansion
 *********************************************************)

lemma eq_Prefix_step: "a -> P =F<ev> ? x:{a} -> P"
apply (simp add: eqsf_decompo)
apply (rule conjI)
apply (simp add: Prefixt_eval Prefixt_def)
apply (simp add: Pre_choicet_eval)

apply (simp add: Prefixf_eval Prefixf_def)
apply (simp add: Pre_choicef_eval)
done

(*********************************************************
                    Ext choice expansion
 *********************************************************)

(*** trace part ***)

lemma eq_Ext_choicet_step: 
  "[[? :X -> Pf [+] ? :Y -> Qf]]T ev =
   [[Ext_choice_step X Y Pf Qf]]T ev"
apply (simp add: Ext_choice_step_def)
apply (simp add: Ext_choicet_eval)
apply (simp add: Pre_choicet_eval)
apply (simp add: IFt_eval Int_choicet_eval)
apply (auto)
done

(*** failure part ***)

(* subgoal 1 *)

lemma eq_Ext_choicef_step_subgoal1: 
   "a = []t & Ev ` X Int b = {} & Ev ` Y Int b = {}
    ==> Ev ` (X Un Y) Int b = {}"
apply (auto)
done

(* subgoal 4 *)

lemma eq_Ext_choicef_step_subgoal4: 
   "[| a = []t ; Ev ` (X Un Y) Int b = {} |] ==>
       Ev ` X Int b = {} & Ev ` Y Int b = {}"
apply (auto)
done

(*** main ***)

lemma eq_Ext_choicef_step: 
  "[[? :X -> Pf [+] ? :Y -> Qf]]F ev =
   [[Ext_choice_step X Y Pf Qf]]F ev"
apply (simp add: Ext_choice_step_def)
apply (simp add: Ext_choicef_eval)
apply (simp add: tick_notin_Pre_choicet)
apply (simp add: Pre_choicef_eval)
apply (simp add: IFf_eval Int_choicef_eval)

apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: nil_in_trace event_app_in_trace Rep_trace Abs_trace_inject)

apply (rule equalityI)

      (* <= *)

apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (rule conjI)

      (* <= sub 1 *)

apply (intro impI)
apply (rule disjI1)
apply (simp add: eq_Ext_choicef_step_subgoal1)

      (* <= sub 2 *)

apply (intro impI)
apply (erule conjE)+
apply (erule disjE)
apply (simp)
apply (erule disjE)
apply (elim conjE exE)
apply (rule_tac x="aa" in exI)
apply (simp)

apply (rule conjI)
apply (intro impI)
apply (rule_tac x="s" in exI)
apply (simp)
apply (intro impI)
apply (rule_tac x="s" in exI)
apply (simp)

      (* <= sub 3 *)

apply (simp)
apply (elim conjE exE)
apply (rule_tac x="aa" in exI)
apply (simp)

apply (rule conjI)
apply (intro impI)
apply (rule_tac x="s" in exI)
apply (simp)
apply (intro impI)
apply (rule_tac x="s" in exI)
apply (simp)

      (* => *)

apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (rule conjI)

      (* => sub 4 *)

apply (intro impI)
apply (rule eq_Ext_choicef_step_subgoal4)
apply (simp)
apply (simp)

      (* => sub 5 *)

apply (intro impI)
apply (elim conjE exE)

apply (case_tac "aa : X")
apply (case_tac "aa : Y")

         (* aa : X; aa : Y *)

apply (simp)
apply (elim conjE exE)
apply (rule disjI2)
apply (simp add: nil_in_trace event_app_in_trace Rep_trace Abs_trace_inject)

             (* (s, b) : [[P]]F ev *)

apply (erule disjE)
apply (rule disjI1)
apply (rule_tac x="s" in exI)
apply (simp)

             (* (s, b) : [[Q]]F ev *)

apply (rule disjI2)
apply (rule_tac x="s" in exI)
apply (simp)

         (* aa : X; aa ~: Y *)

apply (simp)
apply (elim conjE exE)
apply (rule disjI2)
apply (simp add: nil_in_trace event_app_in_trace Rep_trace Abs_trace_inject)

apply (rule_tac x="s" in exI)
apply (simp)

         (* aa ~: X *)

apply (simp)
apply (elim conjE exE)
apply (rule disjI2)
apply (simp add: nil_in_trace event_app_in_trace Rep_trace Abs_trace_inject)

apply (rule_tac x="s" in exI)
apply (simp)
done

(*** proc ***)

lemma eq_Ext_choice_step: 
  "(? :X -> Pf) [+] (? :Y -> Qf) =F<ev> 
    Ext_choice_step X Y Pf Qf"
apply (simp add: eqsf_decompo)
apply (simp add: eq_Ext_choicet_step)
apply (simp add: eq_Ext_choicef_step)
done

(*********************************************************
                    Parallel expansion
 *********************************************************)

(*** trace part (subset) ***)

lemma eq_Parallelt_step_subset: 
  "[[? :Y -> Pf |[X]| ? :Z -> Qf]]T ev <=
   [[Parallel_step X Y Z Pf Qf]]T ev"
apply (simp add: Parallel_step_def)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (elim conjE exE)

apply (insert trace_nil_or_Tick_or_Ev)
apply (drule_tac x="x" in spec)

          (* x = []t *)

apply (erule disjE)
apply (simp add: par_trace_nil)
apply (simp add: Pre_choicet_eval)

          (* x = [Tick]t --> contradict *)

apply (erule disjE)
apply (simp add: par_trace_Tick)
apply (simp add: tick_notin_Pre_choicet)

          (* x = [Ev a]t @t u *)

apply (elim conjE exE)
apply (rename_tac x st s t a u)

apply (insert par_trace_head_only_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: trace_event_app_in_trace)

              (* a : X *)

apply (erule disjE)
apply (elim conjE exE)
apply (simp add: Pre_choicet_eval)

apply (simp add: event_app_not_nil)
apply (elim conjE exE)
apply (simp)

apply (rule_tac x="a" in exI)
apply (rule_tac x="u" in exI)
apply (simp)

apply (erule same_head_tailE)
apply (erule same_head_tailE)
apply (simp)

apply (simp add: IFt_eval)
apply (erule event_in_Ev_XE)
apply (simp)

apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="s' |[X]|set t'" in exI)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: Rep_trace_inject)

              (* a ~: X & left *)

apply (erule conjE)
apply (erule disjE)
apply (elim conjE exE)
apply (simp add: Pre_choicet_eval)

apply (simp add: event_app_not_nil)
apply (elim conjE exE)
apply (simp)

apply (rule_tac x="a" in exI)
apply (rule_tac x="u" in exI)
apply (simp)

apply (erule same_head_tailE)
apply (simp add: IFt_eval)
apply (erule event_notin_Ev_XE)
apply (simp)

           (* t = []t *)

apply (erule disjE)
apply (rule conjI)

                (* *)

apply (intro impI)
apply (simp add: Int_choicet_eval)
apply (rule disjI1)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="s' |[X]|set []t" in exI)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="[]t" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicet_eval)

                (* *)

apply (intro impI)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="s' |[X]|set []t" in exI)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="[]t" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicet_eval)

           (* EX a s. t = [Ev a]t @t s & a : Z & s : [[Q]]T ev *)

apply (rule conjI)

                (* *)

apply (intro impI)
apply (simp add: Int_choicet_eval)
apply (rule disjI1)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="s' |[X]|set t" in exI)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicet_eval)

                (* *)

apply (intro impI)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="s' |[X]|set t" in exI)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicet_eval)

              (* a ~: X & right *)

apply (elim conjE exE)
apply (simp add: Pre_choicet_eval)

apply (simp add: event_app_not_nil)
apply (elim conjE exE)
apply (simp)

apply (rule_tac x="a" in exI)
apply (rule_tac x="u" in exI)
apply (simp)

apply (erule same_head_tailE)
apply (simp add: IFt_eval)
apply (erule event_notin_Ev_XE)
apply (simp)

           (* t = []t *)

apply (erule disjE)
apply (rule conjI)

                (* *)

apply (intro impI)
apply (simp add: Int_choicet_eval)
apply (rule disjI2)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="[]t |[X]|set t'" in exI)
apply (simp)
apply (rule_tac x="[]t" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicet_eval)

                (* *)

apply (intro impI)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="[]t |[X]|set t'" in exI)
apply (simp)
apply (rule_tac x="[]t" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicet_eval)

           (* t = [Ev ab]t @t sb & ab : Z; sb : [[Q]]T ev *)

apply (rule conjI)

                (* *)

apply (intro impI)
apply (simp add: Int_choicet_eval)
apply (rule disjI2)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="s |[X]|set t'" in exI)
apply (simp)
apply (rule_tac x="s" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicet_eval)

                (* *)

apply (intro impI)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)

apply (rule_tac x="s |[X]|set t'" in exI)
apply (simp)
apply (rule_tac x="s" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicet_eval)
done

(*** trace part (supset) ***)

lemma eq_Parallelt_step_supset: 
  "[[Parallel_step X Y Z Pf Qf]]T ev <=
   [[? :Y -> Pf |[X]| ? :Z -> Qf]]T ev"
apply (simp add: Parallel_step_def)
apply (simp add: Pre_choicet_eval)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)
apply (simp add: subset_def Ball_def)

      (* []t *)

apply (rule conjI)
apply (rule_tac x="[]t |[X]|set []t" in exI)
apply (simp add: par_trace_nil)
apply (rule_tac x="[]t" in exI)
apply (rule_tac x="[]t" in exI)
apply (simp add: Pre_choicet_eval)

      (* [Ev a]t @t s *)

apply (intro allI impI)
apply (elim conjE exE)

          (* a : X & a : Y & a : Z *)

apply (erule disjE)
apply (simp add: IFt_eval)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)
apply (elim conjE exE)
apply (rename_tac x a u st s' t')

apply (rule_tac x="([Ev a]t @t s') |[X]|set ([Ev a]t @t t')" in exI)
apply (rule conjI)
apply (rule_tac x="[Ev a]t @t s'" in exI)
apply (rule_tac x="[Ev a]t @t t'" in exI)
apply (simp add: Pre_choicet_eval event_app_not_nil)
apply (rule conjI)
apply (rule_tac x="a" in exI)
apply (rule_tac x="s'" in exI)
apply (simp)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t'" in exI)
apply (simp)

apply (erule event_in_Ev_X_invE)
apply (insert par_trace_head_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="[Ev a]t @t s'" in spec)
apply (drule_tac x="[Ev a]t @t t'" in spec)
apply (simp)
apply (drule mp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

apply (case_tac "a:Y")
apply (case_tac "a:Z")

          (* a : Y & a ~: X & a : Z *)

apply (simp add: IFt_eval)
apply (simp add: Int_choicet_eval)

                 (* s : [[P |[X]| ? x:Z -> Q]]T ev *)

apply (erule disjE)

apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)
apply (elim conjE exE)
apply (rename_tac x a u st s' t)

apply (rule_tac x="([Ev a]t @t s') |[X]|set t" in exI)
apply (rule conjI)
apply (rule_tac x="[Ev a]t @t s'" in exI)
apply (rule_tac x="t" in exI)
apply (simp add: Pre_choicet_eval event_app_not_nil)
apply (rule_tac x="a" in exI)
apply (rule_tac x="s'" in exI)
apply (simp)

apply (erule event_notin_Ev_X_invE)

      (* insert par_trace_head_if *)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="[Ev a]t @t s'" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
apply (erule conjE)

apply (drule mp)
apply (rule_tac x="s'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

                 (* s : [[? x:Y -> P |[X]| Q]]T ev *)

apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)
apply (elim conjE exE)
apply (rename_tac x a u st s t')

apply (rule_tac x="s |[X]|set ([Ev a]t @t t')" in exI)
apply (rule conjI)
apply (rule_tac x="s" in exI)
apply (rule_tac x="[Ev a]t @t t'" in exI)
apply (simp add: Pre_choicet_eval event_app_not_nil)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t'" in exI)
apply (simp)

apply (erule event_notin_Ev_X_invE)

      (* insert par_trace_head_if *)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[Ev a]t @t t'" in spec)
apply (simp)
apply (erule conjE)

apply (rotate_tac -1)
apply (drule mp)
apply (rule_tac x="t'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

          (* a : Y & a ~: X & a ~: Z *)

apply (simp add: IFt_eval)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)
apply (elim conjE exE)
apply (rename_tac x a u st s' t)

apply (rule_tac x="([Ev a]t @t s') |[X]|set t" in exI)
apply (rule conjI)
apply (rule_tac x="[Ev a]t @t s'" in exI)
apply (rule_tac x="t" in exI)
apply (simp add: Pre_choicet_eval event_app_not_nil)
apply (rule_tac x="a" in exI)
apply (rule_tac x="s'" in exI)
apply (simp)

apply (erule event_notin_Ev_X_invE)

      (* insert par_trace_head_if *)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="[Ev a]t @t s'" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
apply (erule conjE)

apply (drule mp)
apply (rule_tac x="s'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

apply (case_tac "a:Z")

          (* a ~: Y & a ~: X & a : Z *)
apply (simp add: IFt_eval)
apply (simp add: Parallelt_eval)
apply (simp add: Parallelt_def)
apply (elim conjE exE)
apply (rename_tac x a u st s t')

apply (rule_tac x="s |[X]|set ([Ev a]t @t t')" in exI)
apply (rule conjI)
apply (rule_tac x="s" in exI)
apply (rule_tac x="[Ev a]t @t t'" in exI)
apply (simp add: Pre_choicet_eval event_app_not_nil)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t'" in exI)
apply (simp)

apply (erule event_notin_Ev_X_invE)

      (* insert par_trace_head_if *)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[Ev a]t @t t'" in spec)
apply (simp)
apply (erule conjE)

apply (rotate_tac -1)
apply (drule mp)
apply (rule_tac x="t'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

          (* a ~: Y & a ~: X & a ~: Z *)

apply (simp)
done

(*** trace part ***)

lemma eq_Parallelt_step: 
  "[[? :Y -> Pf |[X]| ? :Z -> Qf]]T ev =
   [[Parallel_step X Y Z Pf Qf]]T ev"
apply (rule equalityI)
apply (simp add: eq_Parallelt_step_subset del: Int_iff)
apply (simp add: eq_Parallelt_step_supset del: Int_iff)
done

(*******************************
           failure part
 *******************************)

(*** set empty check for [] ***)

(*** sync ***)

lemma eq_Parallelf_step_subgoal_nil1:
   "[| Ev ` Y Int X1 = {} ; Ev ` Z Int X2 = {} |]
         ==> Ev ` (X Int Y Int Z) Int (X1 Un X2) = {}"
apply (auto)
done

(*** left ***)

lemma eq_Parallelf_step_subgoal_nil2:
   "[| X1 - insert Tick (Ev ` X) = X2 - insert Tick (Ev ` X) ;
         Ev ` Y Int X1 = {} & Ev ` Z Int X2 = {} |]
           ==> Ev ` (Y - X) Int (X1 Un X2) = {}"
apply (rule equalityI)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (simp add: inj_Ev image_Minus_dist_inj)

      (* x ~= Tick *)
apply (insert event_tick_or_Ev)
apply (drule_tac x="x" in spec)
apply (erule disjE)
apply (simp add: image_def)
apply (elim conjE exE)

apply (case_tac "x ~: X1 & x ~: X2")
apply (simp)

         (* Ev a : X1 --> contradict *)

apply (simp)
apply (erule disjE)
apply (blast)

         (* Ev a : X2  --> a:X1 --> contradict *)

apply (simp)
apply (erule equalityE)
apply (rotate_tac -1)
apply (erule subsetE)
apply (simp add: Ball_def)
apply (drule_tac x="Ev a" in spec)
apply (simp)
apply (blast)

apply (simp)
done

(*** right ***)
lemma eq_Parallelf_step_subgoal_nil3:
   "[| X1 - insert Tick (Ev ` X) = X2 - insert Tick (Ev ` X) ;
         Ev ` Y Int X1 = {} & Ev ` Z Int X2 = {} |]
           ==> Ev ` (Z - X) Int (X1 Un X2) = {}"
apply (insert eq_Parallelf_step_subgoal_nil2[of X2 X X1 Z Y])
apply (auto)
done

(*** failure part (subset) ***)

lemma eq_Parallelf_step_subset: 
  "[[? :Y -> Pf |[X]| ? :Z -> Qf]]F ev <=
   [[Parallel_step X Y Z Pf Qf]]F ev"
apply (simp add: Parallel_step_def)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (elim conjE exE)
apply (rename_tac v X12 X1 X2 s t)

apply (insert trace_nil_or_Tick_or_Ev)
apply (drule_tac x="v" in spec)

          (* v = []t *)

apply (erule disjE)
apply (simp add: par_trace_nil)
apply (simp add: Pre_choicef_eval)
apply (rule disjI1)
apply (simp add: event_app_not_nil_sym)
apply (simp add: eq_Parallelf_step_subgoal_nil1
                 eq_Parallelf_step_subgoal_nil2
                 eq_Parallelf_step_subgoal_nil3 
                 image_Un_dist Int_Un_distrib2)

          (* v = [Tick]t --> contradict *)

apply (erule disjE)
apply (simp add: par_trace_Tick)
apply (simp add: tick_notin_Pre_choicef)

          (* v = [Ev a]t @t u *)

apply (elim conjE exE)
apply (rename_tac v X12 X1 X2 s t a u)

apply (insert par_trace_head_only_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: trace_event_app_in_trace)

              (* a : X *)

apply (erule disjE)
apply (elim conjE exE)
apply (simp add: Pre_choicef_eval)

apply (simp add: event_app_not_nil)
apply (elim conjE exE)
apply (simp)

apply (rule_tac x="a" in exI)
apply (rule_tac x="u" in exI)
apply (simp)

apply (erule same_head_tailE)
apply (erule same_head_tailE)
apply (simp)

apply (simp add: IFf_eval)
apply (erule event_in_Ev_XE)
apply (simp)

apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: Rep_trace_inject)

              (* a ~: X & left *)

apply (erule conjE)
apply (erule disjE)
apply (elim conjE exE)
apply (simp add: Pre_choicef_eval)

apply (simp add: event_app_not_nil)
apply (elim conjE exE)
apply (simp)

apply (rule_tac x="a" in exI)
apply (rule_tac x="u" in exI)
apply (simp)

apply (erule same_head_tailE)
apply (simp)

apply (simp add: IFf_eval)
apply (erule event_notin_Ev_XE)
apply (simp)

                (* aa : Z *)

apply (rule conjI)

apply (intro impI)
apply (simp add: Int_choicef_eval)
apply (rule disjI1)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicef_eval)

                (* aa ~: Z *)

apply (intro impI)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicef_eval)

              (* a ~: X & right *)

apply (elim conjE exE)
apply (simp add: Pre_choicef_eval)

apply (simp add: event_app_not_nil)
apply (elim conjE exE)
apply (simp)

apply (rule_tac x="a" in exI)
apply (rule_tac x="u" in exI)
apply (simp)

apply (erule same_head_tailE)
apply (simp)

apply (simp add: IFf_eval)
apply (erule event_notin_Ev_XE)
apply (simp)

                (* aa : Y *)

apply (rule conjI)

apply (intro impI)
apply (simp add: Int_choicef_eval)
apply (rule disjI2)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="s" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicef_eval)

                (* aa ~: Z *)

apply (intro impI)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="s" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: Rep_trace_inject)
apply (simp add: Pre_choicef_eval)
done

(*** failure part (supset) ***)

(*  
               X12 = X' - insert Tick (Ev ` X)
               Y12 = X' Int insert Tick (Ev ` X)
               X1  = X12 Un (Y12 - Y)
               X2  = X12 Un (Y12 - Z)

                       lemmas
*)

lemma eq_Parallelf_step_supset_X':
   "Ev ` (X Int Y Int Z Un (Y - X) Un (Z - X)) Int X' = {}
         ==> X' = X' - insert Tick (Ev ` X) Un
                 (X' Int insert Tick (Ev ` X) - Ev ` Y) Un
                 (X' - insert Tick (Ev ` X) Un
                 (X' Int insert Tick (Ev ` X) - Ev ` Z))"
apply (auto)
done

lemma eq_Parallelf_step_supset_YZ:
   "X' - insert Tick (Ev ` X) Un
        (X' Int insert Tick (Ev ` X) - Ev ` Y) -
         insert Tick (Ev ` X) =
    X' - insert Tick (Ev ` X) Un
        (X' Int insert Tick (Ev ` X) - Ev ` Z) -
         insert Tick (Ev ` X)"
apply (auto)
done

lemma eq_Parallelf_step_supset_in_Y:
   "Ev ` (X Int Y Int Z Un (Y - X) Un (Z - X)) Int X' = {}
         ==> Ev ` Y Int ((X' - insert Tick (Ev ` X)) 
                         Un (X' Int insert Tick (Ev ` X) - Ev ` Y)) = {}"
apply (auto)
done

lemma eq_Parallelf_step_supset_in_Z:
   "Ev ` (X Int Y Int Z Un (Y - X) Un (Z - X)) Int X' = {}
         ==> Ev ` Z Int ((X' - insert Tick (Ev ` X)) 
                         Un (X' Int insert Tick (Ev ` X) - Ev ` Z)) = {}"
apply (auto)
done

(*** main part ***)

lemma eq_Parallelf_step_supset: 
  "[[Parallel_step X Y Z Pf Qf]]F ev <=
   [[? :Y -> Pf |[X]| ? :Z -> Qf]]F ev"
apply (simp add: Parallel_step_def)
apply (simp add: Pre_choicef_eval)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)
apply (simp add: subset_def Ball_def)

apply (intro allI)
apply (rule conjI)

      (* v = []t *)

apply (intro impI)
apply (rename_tac v X')

apply (rule_tac x="(X' - insert Tick (Ev ` X)) Un
                   ((X' Int insert Tick (Ev ` X)) - Ev ` Y)" in exI)
apply (rule_tac x="(X' - insert Tick (Ev ` X)) Un
                   ((X' Int insert Tick (Ev ` X)) - Ev ` Z)" in exI)
apply (simp add: eq_Parallelf_step_supset_X')
apply (simp add: eq_Parallelf_step_supset_YZ)
apply (rule_tac x="[]t" in exI)
apply (rule_tac x="[]t" in exI)
apply (simp add: par_trace_nil)
apply (simp add: Pre_choicef_eval)
apply (erule conjE)
apply (simp add: eq_Parallelf_step_supset_in_Y
                 eq_Parallelf_step_supset_in_Z)

      (* v = [Ev a]t @t u *)

apply (intro impI)
apply (elim conjE exE)
apply (rename_tac v X' a u)

          (* a : X & a : Y & a : Z *)

apply (erule disjE)
apply (simp add: IFf_eval)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)
apply (elim conjE exE)
apply (rename_tac v X' a u X1 X2 s' t')

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="[Ev a]t @t s'" in exI)
apply (rule_tac x="[Ev a]t @t t'" in exI)

apply (erule event_in_Ev_X_invE)
apply (insert par_trace_head_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="[Ev a]t @t s'" in spec)
apply (drule_tac x="[Ev a]t @t t'" in spec)
apply (simp)
apply (drule mp)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

apply (simp add: Pre_choicef_eval)
apply (rule conjI)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="s'" in exI)
apply (simp)

apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t'" in exI)
apply (simp)

      (* to avode infinite simp this is removed here *)
apply (erule par_trace_head_if_remE)

          (* a ~: X  *)

apply (case_tac "a:Y")
apply (case_tac "a:Z")

              (* a : Y & a : Z *) 

apply (simp add: IFf_eval)
apply (simp add: Int_choicef_eval)

                   (* (u, X') : [[P |[X]| ? x:Z -> Q]]F ev *)

apply (erule disjE)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)
apply (elim conjE exE)
apply (rename_tac v X' a u X1 X2 s' t)

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="[Ev a]t @t s'" in exI)
apply (rule_tac x="t" in exI)

apply (erule event_notin_Ev_X_invE)
apply (insert par_trace_head_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="[Ev a]t @t s'" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (erule conjE)
apply (drule mp)
apply (rule_tac x="s'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

apply (simp add: Pre_choicef_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="s'" in exI)
apply (simp)

                   (* (u, X') : [[? x:Y -> P |[X]| Q]]F ev *)

apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)
apply (elim conjE exE)
apply (rename_tac v X' a u X1 X2 s t')

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="s" in exI)
apply (rule_tac x="[Ev a]t @t t'" in exI)

apply (erule event_notin_Ev_X_invE)

apply (insert par_trace_head_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[Ev a]t @t t'" in spec)
apply (simp)

apply (erule conjE)
apply (rotate_tac -1)
apply (drule mp)
apply (rule_tac x="t'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

apply (simp add: Pre_choicef_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t'" in exI)
apply (simp)

      (* to avode infinite simp this is removed here *)
apply (erule par_trace_head_if_remE)+

              (* a : Y & a ~: Z *) 

apply (simp add: IFf_eval)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)
apply (elim conjE exE)
apply (rename_tac v X' a u X1 X2 s' t)

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="[Ev a]t @t s'" in exI)
apply (rule_tac x="t" in exI)

apply (erule event_notin_Ev_X_invE)
apply (insert par_trace_head_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="[Ev a]t @t s'" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (erule conjE)
apply (drule mp)
apply (rule_tac x="s'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

apply (simp add: Pre_choicef_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="s'" in exI)
apply (simp)

      (* to avode infinite simp this is removed here *)
apply (erule par_trace_head_if_remE)+

apply (case_tac "a:Z")

              (* a ~: Y & a : Z *) 

apply (simp add: IFf_eval)
apply (simp add: Parallelf_eval)
apply (simp add: Parallelf_def)
apply (elim conjE exE)
apply (rename_tac v X' a u X1 X2 s t')

apply (rule_tac x="X1" in exI)
apply (rule_tac x="X2" in exI)
apply (simp)
apply (rule_tac x="s" in exI)
apply (rule_tac x="[Ev a]t @t t'" in exI)

apply (erule event_notin_Ev_X_invE)

apply (insert par_trace_head_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Ev a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[Ev a]t @t t'" in spec)
apply (simp)

apply (erule conjE)
apply (rotate_tac -1)
apply (drule mp)
apply (rule_tac x="t'" in exI)
apply (simp add: trace_event_app_in_trace)
apply (simp)

apply (simp add: Pre_choicef_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t'" in exI)
apply (simp)

              (* a ~: Y & a ~: Z --> contradiction *) 
apply (simp)
done

(*** failure part ***)

lemma eq_Parallelf_step: 
  "[[? :Y -> Pf |[X]| ? :Z -> Qf]]F ev =
   [[Parallel_step X Y Z Pf Qf]]F ev"
apply (rule equalityI)
apply (simp add: eq_Parallelf_step_subset del: Int_iff)
apply (simp add: eq_Parallelf_step_supset del: Int_iff)
done

(*** proc ***)

lemma eq_Parallel_step:
  "? :Y -> Pf |[X]| ? :Z -> Qf
       =F<ev>
       Parallel_step X Y Z Pf Qf"
apply (simp add: eqsf_decompo)
apply (simp add: eq_Parallelt_step)
apply (simp add: eq_Parallelf_step)
done

(*********************************************************
                      Hide expansion
 *********************************************************)

(*** trace part (subset) ***)

lemma eq_Hidet_step_subset: 
  "[[(? :Y -> Pf) -- X]]T ev <= [[Hide_step X Y Pf]]T ev"
apply (simp add: Hide_step_def)
apply (simp add: Hidet_eval)
apply (simp add: Hidet_def)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (elim conjE exE)
apply (rename_tac s t)

apply (simp add: IFt_eval)
apply (rule conjI)

      (* Y Int X = {} *)

apply (intro impI)
apply (simp add: Pre_choicet_eval)
apply (erule disjE)

         (* t = []t *)

apply (simp add: hide_trace_nil)

         (* t = [Ev a]t @t t'; a : Y; t' : [[P]]T ev *)

apply (elim conjE exE)
apply (rename_tac s t a t')
apply (rule disjI2)
apply (simp add: Hidet_eval)
apply (simp add: Hidet_def)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t' --tr X" in exI)

apply (case_tac "a : X")

             (* a:X --> contradiction *)  
apply (blast)
             (* therefore a ~: X *)

apply (erule event_notin_Ev_X_invE)
apply (simp add: hide_trace_head_notin trace_event_app_in_trace)
apply (rule_tac x="t'" in exI)
apply (simp)

      (* Y Int X ~= {} *)

apply (intro impI)
apply (simp add: Pre_choicet_eval)
apply (erule disjE)

         (* t = []t *)

apply (simp add: hide_trace_nil)
apply (simp add: Timeoutt_eval)
apply (simp add: eval_T_in nil_t_in_all_dom)

         (* t = [Ev a]t @t t'; a : Y; t' : [[P]]T ev *)

apply (elim conjE exE)
apply (rename_tac s t a t')

apply (case_tac "a : X")

             (* a:X *)

apply (simp add: Timeoutt_eval)
apply (rule disjI2)

apply (rotate_tac -1)
apply (erule event_in_Ev_X_invE)
apply (simp add: hide_trace_head_in)
apply (simp add: Int_bindt_eval)
apply (rule disjI2)
apply (rule_tac x="[[Pf a -- X]]T ev" in exI)
apply (rule conjI)
apply (rule_tac x="a" in exI)
apply (simp)

apply (simp add: Hidet_eval)
apply (simp add: Hidet_def)
apply (rule_tac x="t'" in exI)
apply (simp)

             (* a~:X *)

apply (simp add: Timeoutt_eval)
apply (rule disjI1)

apply (rotate_tac -1)
apply (erule event_notin_Ev_X_invE)
apply (simp add: hide_trace_head_notin trace_event_app_in_trace)

apply (simp add: Pre_choicet_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t' --tr X" in exI)
apply (simp)

apply (simp add: Hidet_eval)
apply (simp add: Hidet_def)
apply (rule_tac x="t'" in exI)
apply (simp)
done

(*** trace part (supset) ***)

lemma eq_Hidet_step_supset: 
  "[[Hide_step X Y Pf]]T ev <= [[(? :Y -> Pf) -- X]]T ev"
apply (simp add: Hide_step_def)
apply (case_tac "Y Int X = {}")

      (* Y Int X = {} *)

apply (simp add: IFt_eval)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (simp add: Pre_choicet_eval)
apply (erule disjE)

         (* x = []t *)

apply (simp add: Hidet_eval)
apply (simp add: Hidet_def)
apply (rule_tac x="[]t" in exI)
apply (simp add: hide_trace_nil)
apply (simp add: eval_T_in nil_t_in_all_dom)

         (* s = [Ev a]t @t s'; a : Y; s' : [[P]]T ev *)

apply (elim conjE exE)
apply (rename_tac s a s')

apply (simp add: Hidet_eval)
apply (simp add: Hidet_def)
apply (elim conjE exE)

apply (case_tac "a : X")

             (* a:X --> contradiction *)  
apply (blast)
             (* therefore a ~: X *)

apply (erule event_notin_Ev_X_invE)
apply (rule_tac x="[Ev a]t @t t" in exI)
apply (simp add: hide_trace_head_notin trace_event_app_in_trace)

apply (simp add: Pre_choicet_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t" in exI)
apply (simp)

      (* Y Int X ~= {} *)

apply (simp add: IFt_eval)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)

apply (simp add: Timeoutt_eval)
apply (erule disjE)

          (* x : [[? x:(Y - X) -> P -- X]]T ev *)

apply (simp add: Pre_choicet_eval)
apply (erule disjE)

              (* x = []t *)

apply (simp add: eval_T_in nil_t_in_all_dom)

              (* s = [Ev a]t @t s'; a : Y; a ~: X;  s' : [[P -- X]]T ev  *)

apply (elim conjE exE)
apply (rename_tac s a s')

apply (simp add: Hidet_eval)
apply (simp add: Hidet_def)
apply (elim conjE exE)
apply (rule_tac x="[Ev a]t @t t" in exI)
apply (erule event_notin_Ev_X_invE)
apply (simp add: hide_trace_head_notin trace_event_app_in_trace)

apply (simp add: Pre_choicet_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t" in exI)
apply (simp)

          (* x : [[(? x:Y -> P) -- X]]T ev *)

apply (simp add: Int_bindt_eval)
apply (erule disjE)

              (* x = []t *)

apply (simp add: eval_T_in nil_t_in_all_dom)

              (* x : [[P -- X]]T ev; a : Y; a : X *)

apply (elim conjE exE)
apply (simp add: Hidet_eval)
apply (simp add: Hidet_def)
apply (elim conjE exE)

apply (rule_tac x="[Ev a]t @t t" in exI)
apply (rotate_tac 2)
apply (erule event_in_Ev_X_invE)
apply (simp add: hide_trace_head_in)

apply (simp add: Pre_choicet_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t" in exI)
apply (simp)
done

(*** trace part ***)

lemma eq_Hidet_step:
  "[[(? :Y -> Pf) -- X]]T ev = [[Hide_step X Y Pf]]T ev"
apply (rule equalityI)
apply (simp add: eq_Hidet_step_subset)
apply (simp add: eq_Hidet_step_supset)
done

(*** failure part (subset) ***)

lemma eq_Hidef_step_subset: 
  "[[(? :Y -> Pf) -- X]]F ev <= [[Hide_step X Y Pf]]F ev"
apply (simp add: Hide_step_def)
apply (simp add: Hidef_eval)
apply (simp add: Hidef_def)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (elim conjE exE)
apply (rename_tac s Z t)

apply (simp add: IFf_eval)
apply (rule conjI)

      (* Y Int X = {} *)

apply (intro impI)
apply (simp add: Pre_choicef_eval)
apply (erule disjE)

         (* t = []t *)

apply (rule disjI1)
apply (simp add: hide_trace_nil)
apply (simp add: Int_Un_distrib)

         (* EX a s. t = [Ev a]t @t s & a : Y & (s, Ev ` X Un Z) : [[P]]F ev *)

apply (rule disjI2)
apply (elim conjE exE)
apply (rename_tac s Z t a t')
apply (rule_tac x="a" in exI)
apply (rule_tac x="t' --tr X" in exI)

apply (case_tac "a : X")

             (* a:X --> contradiction *)  
apply (blast)
             (* therefore a ~: X *)

apply (erule event_notin_Ev_X_invE)
apply (simp add: hide_trace_head_notin trace_event_app_in_trace)
apply (simp add: Hidef_eval)
apply (simp add: Hidef_def)
apply (rule_tac x="t'" in exI)
apply (simp)

      (* Y Int X ~= {} *)

apply (intro impI)
apply (simp add: Pre_choicef_eval)
apply (erule disjE)

         (* t = []t --> contradiction *)

apply (simp add: Int_Un_distrib)
apply (blast)

         (* EX a s. t = [Ev a]t @t s & a : Y & (s, Ev ` X Un Z) : [[P]]F ev *)

apply (elim conjE exE)
apply (rename_tac s Z t a t')

apply (case_tac "a : X")

             (* a:X *)

apply (simp add: Timeoutf_eval)
apply (rule disjI1)

apply (simp add: Int_bindf_eval)
apply (rule_tac x="[[Pf a -- X]]F ev" in exI)
apply (rule conjI)
apply (rule_tac x="a" in exI)
apply (simp)

apply (rotate_tac -1)
apply (erule event_in_Ev_X_invE)
apply (simp add: hide_trace_head_in)

apply (simp add: Hidef_eval)
apply (simp add: Hidef_def)
apply (rule_tac x="t'" in exI)
apply (simp)

             (* a~:X *)

apply (simp add: Timeoutf_eval)
apply (rule disjI2)
apply (rule disjI1)

apply (erule event_notin_Ev_X_invE)
apply (simp add: hide_trace_head_notin trace_event_app_in_trace)
apply (simp add: event_app_not_nil)

apply (simp add: Pre_choicef_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t' --tr X" in exI)

apply (simp add: Hidef_eval)
apply (simp add: Hidef_def)
apply (rule_tac x="t'" in exI)
apply (simp)
done

(*** failure part (supset) ***)

lemma eq_Hidef_step_supset: 
  "[[Hide_step X Y Pf]]F ev <= [[(? :Y -> Pf) -- X]]F ev"
apply (simp add: Hide_step_def)
apply (case_tac "Y Int X = {}")

      (* Y Int X = {} *)

apply (simp add: IFf_eval)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (rename_tac s Z)
apply (simp add: Pre_choicef_eval)

apply (erule disjE)

        (* s = []t & Ev ` Y Int Z = {} *)

apply (simp add: Hidef_eval)
apply (simp add: Hidef_def)
apply (rule_tac x="[]t" in exI)
apply (simp add: hide_trace_nil)
apply (simp add: Pre_choicef_eval)
apply (blast)

        (* (EX a sa. s = [Ev a]t @t sa & a : Y & (sa, Z) : [[P -- X]]F ev) *)

apply (elim conjE exE)
apply (rename_tac s Z a s')
apply (simp add: Hidef_eval)
apply (simp add: Hidef_def)
apply (elim conjE exE)
apply (rule_tac x="[Ev a]t @t t" in exI)

apply (case_tac "a : X")
             (* a:X --> contradiction *)  
apply (blast)
             (* therefore a ~: X *)

apply (erule event_notin_Ev_X_invE)
apply (simp add: hide_trace_head_notin trace_event_app_in_trace)

apply (simp add: Pre_choicef_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t" in exI)
apply (simp)

      (* Y Int X ~= {} *)

apply (simp add: IFf_eval)
apply (simp add: subset_def Ball_def)
apply (intro allI impI)
apply (rename_tac s Z)
apply (simp add: Timeoutf_eval)

        (* (s, Z) : [[! x:(Y Int X) .. P -- X]]F ev *)

apply (erule disjE)
apply (simp add: Int_bindf_eval)
apply (elim conjE exE)

apply (simp add: Hidef_eval)
apply (simp add: Hidef_def)
apply (elim conjE exE)

apply (rule_tac x="[Ev a]t @t t" in exI)
apply (rotate_tac 2)
apply (erule event_in_Ev_X_invE)
apply (simp add: hide_trace_head_in)

apply (simp add: Pre_choicef_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t" in exI)
apply (simp)

        (* s ~= []t & (s, Z) : [[? x:(Y - X) -> P -- X]]F ev *)

apply (erule disjE)
apply (simp add: Pre_choicef_eval)
apply (erule conjE)
apply (simp)
apply (elim conjE exE)
apply (rename_tac s Z a s')

apply (simp add: Hidef_eval)
apply (simp add: Hidef_def)
apply (elim conjE exE)

apply (rule_tac x="[Ev a]t @t t" in exI)
apply (erule event_notin_Ev_X_invE)
apply (simp add: hide_trace_head_notin trace_event_app_in_trace)
apply (simp add: Pre_choicef_eval)
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="t" in exI)
apply (simp)

        (* s = []t & Z <= Evset & [Tick]t:[[? x:(Y - X) -> P -- X]]T ev *)

apply (simp add: tick_notin_Pre_choicet)
done

(*** failure part ***)

lemma eq_Hidef_step: 
  "[[(? :Y -> Pf) -- X]]F ev = [[Hide_step X Y Pf]]F ev"
apply (rule equalityI)
apply (simp add: eq_Hidef_step_subset)
apply (simp add: eq_Hidef_step_supset)
done

(*** proc ***)

lemma eq_Hide_step:
  "(? :Y -> Pf) -- X =F<ev> Hide_step X Y Pf"
apply (simp add: eqsf_decompo)
apply (simp add: eq_Hidet_step)
apply (simp add: eq_Hidef_step)
done

end
