           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Trace = Infra:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite (notick | t = []t)                 *)
(*                                                                     *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(***********************************************************

    Type Definitions
      'a event      : type of events      a,b,...
      'a trace      : type of traces      s,t,...

 ***********************************************************)

datatype 'a event = Ev 'a | Tick

consts
  Evset     :: "'a event set"
  EvsetTick :: "'a event set"

defs
  Evset_def     : "Evset     == {a. a ~= Tick}"
  EvsetTick_def : "EvsetTick == UNIV"

(* EvsetTick seems to be useless, but it sometimes *)
(* makes proofs to be readable.                    *)

(*******************************
    Evset contains all (Ev a)
 *******************************)

lemma Un_Evset : "ALL X. Ev ` X Un Evset = Evset"
by (auto simp add: Evset_def)

(*******************************
         Ev a : Ev ` X
 *******************************)

(*** simp ***)

lemma event_in_Ev_X: "ALL a X. Ev a : Ev ` X  --> a : X"
by auto

lemma event_notin_Ev_X: "ALL a X. Ev a ~: Ev ` X  --> a ~: X"
by auto

(*** erule ***)

lemma event_in_Ev_XE: 
  "[| Ev a : Ev ` X ; [| a : X ; Ev a : Ev ` X |] ==> R |] ==> R"
by auto

lemma event_notin_Ev_XE: 
  "[| Ev a ~: Ev ` X ; [| a ~: X ; Ev a ~: Ev ` X |] ==> R |] ==> R"
by auto

lemma event_in_Ev_X_invE: 
   "[| a : X ; [| a : X ; Ev a : Ev ` X |] ==> R |] ==> R"
by auto

lemma event_notin_Ev_X_invE: 
   "[| a ~: X ; [| a ~: X ; Ev a ~: Ev ` X |] ==> R |] ==> R"
by auto

(*******************************
         Tick or Ev a
 *******************************)

lemma event_tick_or_Ev: "ALL e. e = Tick | (EX a. e = Ev a)"
apply (intro allI)
apply (induct_tac e)
apply (auto)
done

(*******************************
             inj
 *******************************)

lemma inj_Ev: "inj Ev"
apply (simp add: inj_on_def)
done

(***********************************************************
                    Definition of traces
 ***********************************************************)

(* "typedef" requires defined types are not empty as follows: *)

typedef 'a trace = "{s::('a event) list. Tick ~: set (butlast s)}"
apply (rule_tac x ="[]" in exI)
apply (simp)
done

(*
   declare Abs_trace_inverse [simp]
   declare Abs_trace_inject  [simp]
   declare Rep_trace         [simp]
   declare trace_def         [simp]
*)

(***********************************************************)
(*                  operators over traces                  *)
(*                   syntax translations                   *)
(*                                                         *)
(*    memo    (infixr "@t" 65) = ("_ @t _" [66,65] 65)     *)
(*                                                         *)
(***********************************************************)

syntax
  "_sett"    :: "'a trace => 'a event set"          ("sett")
  "_lengtht" :: "'a trace => nat"                   ("lengtht")
  "_tick"    :: "'a trace => bool"                  ("tick")
  "_notick"  :: "'a trace => bool"                  ("notick")
  "_apprep"  :: "'a trace => 'a trace => 'a list"   (infixr "@rep" 65)
  "_appt"    :: "'a trace => 'a trace => 'a trace"  (infixr "@t" 65)

  "_nil_t"   :: "'a trace"                          ("[]t")
  "@trace"   :: "args => 'a trace"                  ("[(_)]t")

translations
  "sett s"    == "set (Rep_trace s)"
  "lengtht s" == "length (Rep_trace s)"
  "tick s"    == "Tick : sett s"
  "notick s"  == "Tick ~: sett s"
  "s @rep t"  == "Rep_trace s @ Rep_trace t"
  "s @t t"    == "Abs_trace (s @rep t)"

  "[]t"       == "Abs_trace []"
  "[s]t"      == "Abs_trace [s]"

(***********************************************************
                          lemmas
 ***********************************************************)

(*******************************
              ALL
 *******************************)

lemma ALL_Rep_trace: "ALL s. Rep_trace s : trace"
apply (simp add: Rep_trace)
done

(*******************************
      Rep and Abs lemmas
 *******************************)

lemma Rep_trace_eq: 
  "ALL t s. Rep_trace t = s --> s : trace"
apply (simp add: Rep_trace)
done

lemma Abs_inject_inv: 
  "ALL t s. (t : trace & s : trace) --> (s = t) = (Abs_trace s = Abs_trace t)"
apply (simp add: Abs_trace_inject)
done

lemma Abs_Rep_t_inv: 
  "ALL t s. s : trace --> ((Rep_trace t = s) = (t = Abs_trace s))"
apply (intro allI impI)
apply (insert Abs_inject_inv)
apply (drule_tac x="s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
done

lemma Abs_Rep_t_inv_imp: "ALL t s. Rep_trace t = s --> t = Abs_trace s"
apply (intro allI impI)
apply (insert Abs_Rep_t_inv)
apply (drule_tac x="t" in spec)
apply (drule_tac x="s" in spec)
apply (simp add: Rep_trace_eq)
done

lemma Abs_Rep_t: 
  "ALL t s. t:trace --> ((Abs_trace t = s) = (t = Rep_trace s))"
apply (intro allI impI)
apply (insert Abs_Rep_t_inv)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (auto)
done

lemma t_in_Abs_set : 
  "ALL t X. (ALL s:X. s:trace) &  t:Abs_trace ` X --> Rep_trace t : X"
apply (intro allI impI)
apply (simp add: image_def Bex_def Ball_def)
apply (erule conjE)
apply (erule exE)
apply (drule_tac x="x" in spec)
apply (simp add: Abs_trace_inverse)
done

lemma t_in_Rep_set : 
  "ALL t X.  Rep_trace t : X --> t : Abs_trace ` X"
apply (intro allI impI)
apply (simp add: image_def Bex_def Ball_def)
apply (rule_tac x="Rep_trace t" in exI)
apply (simp add: Rep_trace_inverse)
done

lemma ALL_Abs_inverse_inv:
  "ALL s. s:trace --> s = Rep_trace (Abs_trace s)"
apply (simp add: Abs_trace_inverse)
done

lemma ALL_Rep_inverse_inv: "ALL s. s = Abs_trace (Rep_trace s)"
apply (simp add: Rep_trace_inverse)
done

(*******************************
         check in trace
 *******************************)

lemma nil_in_trace : "[]:trace"
apply (simp add: trace_def)
done

lemma tick_in_trace : "[Tick]:trace"
apply (simp add: trace_def)
done

lemma event_in_trace : "ALL (a::'a event). [a]:trace"
apply (simp add: trace_def)
done

lemma event_app_in_trace : "ALL a s. s:trace --> (Ev a)#s:trace"
apply (simp add: trace_def)
done

lemma notick_app_in_trace : "ALL x s. x ~= Tick & s:trace --> x#s:trace"
apply (auto simp add: trace_def)
done

lemma event_app_in_trace_inv : "ALL a s. a#s:trace --> s:trace"
apply (simp add: trace_def)
done

lemma trace_event_app_in_trace: "ALL a s. [Ev a]t @rep s : trace"
apply (intro allI)
apply (insert event_app_in_trace)
apply (drule_tac x="a" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (simp add: event_in_trace Abs_trace_inverse Rep_trace)
done

lemma event_app_in_trace_invI : "a#s:trace ==> s:trace"
apply (insert event_app_in_trace_inv)
apply (drule_tac x="a" in spec)
apply (drule_tac x="s" in spec)
apply (simp)
done

lemma app_event_notick : 
  "ALL s a. Tick ~: set s --> s @ [a]:trace"
apply (simp add: trace_def)
done

lemma app_trace_in_trace_inv : "ALL s t. s @ t:trace --> s:trace"
apply (intro allI impI)
apply (simp add: trace_def)
apply (insert notin_butlast_app_inv)
apply (drule_tac x="Tick" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
done

lemma app_trace_in_trace_inv2 : "ALL s t. s @ t:trace --> s:trace & t:trace"
apply (intro allI impI)
apply (insert app_trace_in_trace_inv)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: trace_def)
apply (insert notin_butlast_app_inv2)
apply (drule_tac x="Tick" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
done

lemma app_event_notick_inv : 
  "ALL s a. s @ [a]:trace --> (Tick ~: set s & s:trace)"
apply (intro allI impI)
apply (simp add: trace_def)
apply (insert butlast_subseteq)
apply (drule_tac x="s" in spec)
apply (auto)
done

lemma app_trace_notick : 
  "ALL s t. (Tick ~: set s & t: trace) --> s @ t : trace"
apply (intro allI impI)
apply (simp add: trace_def)
apply (insert list_last_nil_or_unnil)
apply (drule_tac x="t" in spec)
apply (erule disjE)
apply (simp)
apply (insert butlast_subseteq)
apply (drule_tac x="s" in spec)
apply (auto)

apply (insert ALL_butlast_snoc)
 apply (drule_tac x="s" in spec) (* dummy *)
apply (drule_tac x="s @ sa" in spec)
apply (drule_tac x="a" in spec)
apply (simp)
done

lemma Tick_app_nil_in_trace : "ALL s. Tick#s : trace --> s=[]"
apply (simp add: trace_def)
done

(*******************************
            notick
 *******************************)

lemma Ev_app_notick : "ALL a s. notick s --> notick ([Ev a]t @t s)"
apply (intro allI impI)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: event_app_in_trace Abs_trace_inverse Rep_trace)
done

(*** notick Ev ***)

lemma notick_Ev: "ALL a. notick [Ev a]t"
apply (simp add: event_in_trace Abs_trace_inverse) 
done

(*** notick nil ***)

lemma notick_nil: "notick []t"
apply (simp add: nil_in_trace Abs_trace_inverse) 
done

(***********************************************************
  existstency of @t can be checked by (notick s | t = []t)
 ***********************************************************)

(*** if ***)

lemma trace_app_in_trace_if:
  "ALL s t. (notick s | t = []t) --> s @rep t : trace"
apply (intro allI impI)
apply (erule disjE)

apply (insert ALL_Rep_trace)
apply (drule_tac x="t" in spec)
apply (simp add: trace_def notin_butlast_app)
apply (auto simp add: Abs_trace_inverse nil_in_trace)
done

(*** only if ***)

lemma trace_app_in_trace_only_if:
  "ALL s t. ((s @rep t) : trace) --> (notick s | t = []t)"
apply (intro allI impI)

apply (case_tac "notick s | t = []t")
 apply (simp)

      (* otherwise *)
 apply (simp add: trace_def)

 apply (insert non_nil_exists)
 apply (drule_tac x="Rep_trace t" in spec)
 apply (simp add: Abs_Rep_t_inv Rep_trace  nil_in_trace)
 apply (erule exE)+

 apply (insert ALL_butlast_snoc)
 apply (drule_tac x="Rep_trace s @ sa" in spec)
 apply (drule_tac x="a" in spec)
 apply (simp)
done

(*** iff ***)

lemma trace_app_in_trace:
  "ALL s t. ((s @rep t) : trace) = (notick s | t = []t)"
apply (intro allI)
apply (rule iffI)
apply (simp add: trace_app_in_trace_only_if)
apply (simp add: trace_app_in_trace_if)
done

(*** only if (list) ***)

lemma list_app_in_trace_only_if:
  "ALL s t. s:trace & t:trace & s @ t : trace --> (Tick ~: set s | t = [])"
apply (intro allI impI)
apply (insert trace_app_in_trace_only_if)
apply (drule_tac x="Abs_trace s" in spec)
apply (drule_tac x="Abs_trace t" in spec)
apply (simp_all add: nil_in_trace Abs_trace_inverse Abs_trace_inject)
done

(*** if (list) ***)

lemma list_app_in_trace_if:
  "ALL s t. s:trace & t:trace & (Tick ~: set s | t = []) 
       --> s @ t : trace"
apply (intro allI impI)
apply (insert trace_app_in_trace_if)
apply (drule_tac x="Abs_trace s" in spec)
apply (drule_tac x="Abs_trace t" in spec)
apply (erule conjE)+
apply (erule disjE)
apply (simp_all add: nil_in_trace Abs_trace_inverse Abs_trace_inject)
done

(*** iff (list) ***)

lemma list_app_in_trace:
  "ALL s t. s:trace & t:trace -->
      (s @ t : trace = (Tick ~: set s | t = []))"
apply (intro allI impI)
apply (rule iffI)
apply (simp add: list_app_in_trace_only_if)
apply (simp add: list_app_in_trace_if)
done

(***********************************************************
                trace type <--> list type
 ***********************************************************)

lemma trace_rep_nil_t: "Rep_trace []t = []"
apply (simp add: nil_in_trace Abs_trace_inverse)
done

lemma trace_rep_nil_t_inv: "(Rep_trace x = []) = (x = []t)"
apply (simp add: nil_in_trace Abs_Rep_t_inv)
done

lemma notick_app_trace_rep:
  "ALL s t. notick s --> Rep_trace (s @t t) = s @rep t"
apply (simp add: trace_app_in_trace Abs_trace_inverse)
done

lemma trace_app_nil_rep:
  "ALL s. Rep_trace (s @t []t) = Rep_trace s"
apply (simp add: nil_in_trace Rep_trace Abs_trace_inverse)
done

lemma trace_app_nil:
  "ALL s. (s @t []t) = s"
apply (simp add: nil_in_trace Rep_trace Abs_trace_inverse Rep_trace_inverse)
done

lemma trace_nil_app:
  "ALL s. ([]t @t s) = s"
apply (simp add: nil_in_trace Rep_trace Abs_trace_inverse Rep_trace_inverse)
done

(*** @t --> @rep ***)

lemma appt_to_app: 
  "ALL s t u. s @rep t : trace
      --> (u = s @t t) = (Rep_trace u = s @rep t)"
apply (intro allI impI)
apply (rule iffI)
apply (simp add: Abs_trace_inverse)
apply (simp add: Abs_Rep_t_inv)
done

(*** @t --> @rep (weak assumption) ***)

lemma appt_to_app_or: 
  "ALL s t u. (s @rep t : trace | notick s | t = []t)
      --> (u = s @t t) = (Rep_trace u = s @rep t)"
apply (intro allI impI)
apply (erule disjE)
apply (simp add: appt_to_app)

apply (insert trace_app_in_trace)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: appt_to_app)
done

(*** @t --> @rep (weak assumption) (sym) ***)

lemma appt_to_app_or_sym: 
  "ALL s t u. (s @rep t : trace | notick s | t = []t)
      --> (s @t t = u) = (s @rep t = Rep_trace u)"
apply (intro allI impI)
apply (insert appt_to_app_or)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (drule_tac x="u" in spec)
apply (blast intro: sym)
done

(*** rep @t --> @rep ***)

lemma rep_appt_to_apprep: 
  "ALL s t. (s @rep t : trace | notick s | t = []t)
      --> Rep_trace (s @t t) = s @rep t"
apply (intro allI impI)
apply (auto simp add: trace_app_in_trace_if Abs_trace_inverse)
done


(***********************************************************
                   convenient lemmas
 ***********************************************************)

(*** length ***)

lemma length_app_event_Suc:
  "ALL s a. notick s --> lengtht (s @t [a]t) = Suc (lengtht s)"
apply (simp add: trace_app_in_trace_if Abs_trace_inverse) 
apply (simp add: event_in_trace Abs_trace_inverse)
done

(*** app length ***)

lemma length_app_decompo:
  "ALL s t. notick s --> lengtht (s @t t) = lengtht s + lengtht t"
apply (simp add: trace_app_in_trace_if Abs_trace_inverse) 
done

(*** nil ***)

lemma app_t_nil:
  "ALL s t. (s @rep t : trace | notick s | t = []t)
      --> (s @t t = []t) = (s = []t & t = []t)"
apply (intro allI impI)
apply (simp add: appt_to_app_or_sym)
apply (simp add: nil_in_trace Abs_trace_inverse)
apply (simp add: Abs_Rep_t_inv nil_in_trace)
done

(*** head is not tick ***)

lemma head_notick: 
  "ALL e s. 1 <= length s & e#s : trace --> e ~= Tick"
apply (intro allI impI)
apply (insert list_last_nil_or_unnil)
apply (drule_tac x="s" in spec)
apply (erule disjE)
apply (simp)

apply (erule exE)+
apply (auto simp add: trace_def)
done


(*** nil or unnil ***)

lemma trace_nil_or_unnil: "ALL t. (t = []t) | (EX a s. t=[a]t @t s)"
apply (intro allI)
apply (insert list_nil_or_unnil)
apply (drule_tac x="Rep_trace t" in spec)
apply (erule disjE)
apply (simp add: nil_in_trace Abs_Rep_t_inv)

apply (erule exE)+
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="Abs_trace s" in exI)

apply (insert event_app_in_trace_inv)
apply (drule_tac x="a" in spec)
apply (drule_tac x="s" in spec)
apply (insert ALL_Rep_trace)
apply (drule_tac x="t" in spec)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: Abs_Rep_t_inv)
done


(*** trace --> nil or Tick or etc ***)

lemma trace_nil_or_Tick_or_Ev: "ALL t. 
  (t = []t) | t=[Tick]t | (EX a s. t=[Ev a]t @t s)"
apply (intro allI)
apply (insert list_nil_or_unnil)
apply (drule_tac x="Rep_trace t" in spec)
apply (erule disjE)

  (* t = [] *)

apply (simp add: nil_in_trace Abs_Rep_t_inv)

  (* t ~= [] *)

apply (erule exE)+
apply (rule disjI2)

apply (insert list_nil_or_unnil)
apply (drule_tac x="s" in spec)
apply (erule disjE)

  (* s = [] *)

apply (simp)

apply (insert event_tick_or_Ev)
apply (drule_tac x="a" in spec)
apply (erule disjE)

    (* a = Tick *)

apply (simp add: event_in_trace Abs_Rep_t_inv)

    (* a ~= Tick *)

apply (erule exE)
apply (rule disjI2)
apply (rule_tac x="aa" in exI)
apply (rule_tac x="[]t" in exI)

apply (simp add: event_in_trace nil_in_trace Abs_trace_inverse)
apply (simp add: event_in_trace Abs_Rep_t_inv)

  (* s ~= [] *)

apply (rule disjI2)

apply (erule exE)+
apply (drule_tac x="a" in spec)

apply (insert ALL_Rep_trace)
apply (drule_tac x="t" in spec)
apply (simp)

apply (insert head_notick)
apply (drule_tac x="a" in spec)
apply (drule_tac x="aa#sa" in spec)
apply (simp)

apply (erule exE)+

apply (rule_tac x="ab" in exI)
apply (rule_tac x="Abs_trace (aa # sa)" in exI)

apply (insert event_app_in_trace_inv)
apply (drule_tac x="a" in spec)
apply (drule_tac x="aa # sa" in spec)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: Abs_Rep_t_inv)
done

(*** not nil ***)

lemma event_app_not_nil: "ALL a s. [Ev a]t @t s ~= []t"
apply (simp add: event_in_trace nil_in_trace Rep_trace event_app_in_trace
                 Abs_trace_inverse Abs_trace_inject)
done

lemma event_app_not_nil_sym: "ALL a s. []t ~= [Ev a]t @t s"
apply (simp add: event_in_trace nil_in_trace Rep_trace event_app_in_trace
                 Abs_trace_inverse Abs_trace_inject)
done


(*** same head and tail ***)

lemma same_head_tail: "ALL a b s t. ([Ev a]t @t s = [Ev b]t @t t)
   --> (a = b) & (Rep_trace s = Rep_trace t)"
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: Rep_trace event_app_in_trace Abs_trace_inject)
done

lemma same_head_tailE: "[| [Ev a]t @t s = [Ev b]t @t t ;
  [| a = b ; Rep_trace s = Rep_trace t |] ==> R |] ==> R"
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: Rep_trace event_app_in_trace Abs_trace_inject)
done

(****** lemmas for Csp_Seq.thy ******)

(*** nil or unnil (last) ***)

lemma trace_last_nil_or_unnil: 
  "ALL t. (t = []t) | (EX a s. notick s & t= s @t [a]t)"
apply (intro allI)
apply (insert list_last_nil_or_unnil)
apply (drule_tac x="Rep_trace t" in spec)
apply (erule disjE)
apply (simp add: nil_in_trace Abs_Rep_t_inv)

apply (erule exE)+
apply (rule disjI2)
apply (rule_tac x="a" in exI)
apply (rule_tac x="Abs_trace s" in exI)

apply (insert app_trace_in_trace_inv)
apply (drule_tac x="s" in spec)
apply (drule_tac x="[a]" in spec)
apply (insert ALL_Rep_trace)
apply (drule_tac x="t" in spec)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: Abs_Rep_t_inv)

apply (simp add: trace_def)
done

(*** same last ***)

lemma app_trace_same_last : 
  "ALL s a t b. notick s & notick t & s @t [a]t = t @t [b]t 
      --> (s = t) & (a = b)"
apply (intro allI impI)
apply (elim conjE)
apply (simp add: appt_to_app_or_sym)
apply (simp add: rep_appt_to_apprep)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: Rep_trace_inject)
done

(*** notick & notick -> notick ***)

lemma app_trace_notick_notick:
  "ALL s t. notick s & notick t --> notick (s @t t)"
apply (intro allI impI)
apply (insert trace_app_in_trace_if)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: Abs_trace_inverse)
done

(*** notick & ? -> notick ***)

lemma app_trace_notick_notick_inv:
  "ALL s t. notick s & notick (s @t t) --> notick t"
apply (intro allI impI)
apply (insert trace_app_in_trace_if)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp add: Abs_trace_inverse)
done

(*** app trace associatvity ***)

lemma app_trace_notick_ass :
  "ALL s t u. notick s & notick t --> s @t (t @t u) = (s @t t) @t u"
apply (intro allI impI)
apply (elim conjE)

apply (insert app_trace_notick_notick)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="s @t t" in spec)
apply (drule_tac x="u" in spec)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t @t u" in spec)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="t" in spec)
apply (drule_tac x="u" in spec)

apply (simp add: Abs_trace_inject Abs_trace_inverse)
done

(****************** to add it again ******************)


declare disj_not1 [simp] 



end
