           (*-------------------------------------------*
            |                Csp-Prover                 |
            |                July 2004                  |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Trace_par = Prefix:

(*  The following simplification rules are deleted in this theory file *)
(*  because they unexpectly rewrite (notick | t = []t)                 *)
(*                                                                     *)
(*                  disj_not1: (~ P | Q) = (P --> Q)                   *)

declare disj_not1 [simp del]

(*****************************************************************

         1. s |[X]|list t  : lists  --> list set
         2. s |[X]|set t   : traces --> trace set
         3. 
         4. 

 *****************************************************************)

consts
  par_org :: "'a list * 'a set * 'a list => 'a list set"

recdef par_org "measure (%(s,X,t). (length s)+(length t))"
  "par_org (x#s,X,y#t)
     = (if x:X 
        then (if y:X
              then (if x=y 
                    then {x#v |v. v: par_org (s,X,t)}
                    else {})
              else {y#v |v. v: par_org (x#s,X,t)})
        else (if y:X 
              then {x#v |v. v: par_org (s,X,y#t)}
              else {x#v |v. v: par_org (s,X,y#t)} Un
                   {y#v |v. v: par_org (x#s,X,t)}))"

  "par_org (x#s,X,[])
     = (if x:X then {}
               else {x#v |v. v: par_org (s,X,[])})"

  "par_org ([],X,x#s)
     = (if x:X then {}
               else {x#v |v. v: par_org ([],X,s)})"

  "par_org ([],X,[]) = {[]}"

consts
  par_list :: "'a list => 'a set => 'a list => 'a list set"
                                    ("(_ |[_]|list _)" [76,0,77] 76)
  par_rep  :: "'a trace => 'a set => 'a trace => 'a event list set"
                                    ("(_ |[_]|rep _)" [76,0,77] 76)

defs
  par_list_def : "s |[X]|list t == par_org (s,X,t)"
  par_rep_def : 
    "s |[X]|rep t  == (Rep_trace s) |[(Ev ` X) Un {Tick}]|list (Rep_trace t)"

syntax
  "_par_set" :: "'a trace => 'a set => 'a trace => 'a trace set"
                                    ("(_ |[_]|set _)" [76,0,77] 76)

translations
  "s |[X]|set t" == "Abs_trace ` (s |[X]|rep t)"


(*************************************************************
                    parallel list nil
 *************************************************************)

(*******************************
     par list nil (only if)
 *******************************)

(*** (unnil,unil) ***)

lemma par_list_nil_unnil_unnil:
  "ALL X a s b t. [] ~: (a#s) |[X]|list (b#t)"
apply (auto simp add: par_list_def)
done

(*** (nil,unnil) ***)

lemma par_list_nil_nil_unnil:
  "ALL X b t. [] ~: [] |[X]|list (b#t)"
apply (auto simp add: par_list_def)
done

(*** (unnil,nil) ***)

lemma par_list_nil_unnil_nil:
  "ALL X b t. [] ~: (a#s) |[X]|list []"
apply (auto simp add: par_list_def)
done

(*******************************
     par list nil (only if)
 *******************************)

lemma par_list_nil_only_if:
  "ALL X s t. [] : s |[X]|list t --> s = [] & t = []"
apply (intro allI impI)

apply (insert list_nil_or_unnil)
apply (frule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (erule disjE)
apply (erule disjE)

  (* nil,nil *)

apply (simp)

  (* nil,unnil *)

apply (erule exE)+
apply (simp add: par_list_nil_nil_unnil)

apply (erule disjE)

  (* unnil,nil *)

apply (erule exE)+
apply (simp add: par_list_nil_unnil_nil)

  (* unnil,unnil *)

apply (erule exE)+
apply (simp add: par_list_nil_unnil_unnil)
done

(*******************************
       par list nil (iff)
 *******************************)

lemma par_list_nil:
  "ALL X s t. ([] : s |[X]|list t) = (s = [] & t = [])"
apply (intro allI)
apply (rule iffI)
apply (simp add: par_list_nil_only_if)
apply (simp add: par_list_def)
done


(*************************************************************
                    parallel list head
 *************************************************************)

(*******************************
     par list head (only if)
 *******************************)

(*** (nil,nil) ***)

lemma par_list_nil_nil:
  "ALL X x u. x#u : [] |[X]|list [] = False"
apply (simp add: par_list_def)
done

(*******************************
             Sync
 *******************************)

(*** sync (unnil,unnil) ***)

lemma par_list_sync_unnil_unnil:
  "ALL X x u a s b t. x:X 
     --> (x#u : (a#s) |[X]|list (b#t)) = (a=x & b=x & u : s |[X]|list t)"
apply (auto simp add: par_list_def)
done

(*** sync (nil,unnil) ***)

lemma par_list_sync_nil_unnil: 
  "ALL X x u b t. x:X --> x#u : [] |[X]|list (b#t) = False"
apply (auto simp add: par_list_def)
done

(*** sync (unnil,nil) ***)

lemma par_list_sync_unnil_nil:
  "ALL X x u a s. x:X --> x#u : (a#s) |[X]|list [] = False"
apply (auto simp add: par_list_def)
done

(*** sync (any,any) ***)

lemma par_list_sync_any_any:
  "ALL X x u s t. x:X & x#u : s |[X]|list t
     --> (EX s'. s=x#s' & (EX t'. t=x#t' & u : s' |[X]|list t'))"

apply (intro allI impI)

apply (insert list_nil_or_unnil)
apply (frule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (erule conjE)+
apply (erule disjE)
apply (erule disjE)

  (* nil,nil *)

apply (simp add: par_list_nil_nil)

  (* nil,unnil *)

apply (erule exE)+
apply (simp add: par_list_sync_nil_unnil)

apply (erule disjE)

  (* unnil,nil *)

apply (erule exE)+
apply (simp add: par_list_sync_unnil_nil)

  (* unnil,unnil *)

apply (erule exE)+
apply (simp add: par_list_sync_unnil_unnil)
done

(*******************************
           Interleave
 *******************************)

(*** interleave (unnil,unnil) ***)

lemma par_list_interleave_unnil_unnil:
  "ALL X x u a s b t. x ~: X 
     --> x#u : (a#s) |[X]|list (b#t) 
         = ((a=x & u : s |[X]|list (b#t)) | (b=x & u : (a#s) |[X]|list t))"
apply (auto simp add: par_list_def)
done

(*** interleave (nil,unnil) ***)

lemma par_list_interleave_nil_unnil:
  "ALL X x u b t. x ~: X 
     --> x#u : [] |[X]|list (b#t) = (b=x & u : [] |[X]|list t)"
apply (auto simp add: par_list_def)
done

(*** interleave (unnil,nil) ***)

lemma par_list_interleave_unnil_nil:
  "ALL X x u a s. x ~: X 
     --> x#u : (a#s) |[X]|list [] = (a=x & u : s |[X]|list [])"
apply (auto simp add: par_list_def)
done

(*** interleave (any,any) ***)

lemma par_list_interleave_any_any:
  "ALL X x u s t. x~:X & x#u : s |[X]|list t
     --> (EX s'. s=x#s' & u : s' |[X]|list t ) |
         (EX t'. t=x#t' & u : s  |[X]|list t')"
apply (intro allI impI)

apply (insert list_nil_or_unnil)
apply (frule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (erule conjE)+
apply (erule disjE)
apply (erule disjE)

  (* nil,nil *)

apply (simp add: par_list_nil_nil)

  (* nil,unnil *)

apply (erule exE)+
apply (simp add: par_list_interleave_nil_unnil)

apply (erule disjE)

  (* unnil,nil *)

apply (erule exE)+
apply (simp add: par_list_interleave_unnil_nil)

  (* unnil,unnil *)

apply (erule exE)+
apply (simp add: par_list_interleave_unnil_unnil)
done

(*******************************
    par list head (only if)
 *******************************)

lemma par_list_head_only_if:
  "ALL X x u s t. x#u : s |[X]|list t
     --> (x :X & (EX s'. s=x#s' & (EX t'. t=x#t' & u : s' |[X]|list t'))) |
         (x~:X & ((EX s'. s=x#s' & u : s' |[X]|list t ) |
                  (EX t'. t=x#t' & u : s  |[X]|list t')))"
apply (intro allI impI)

apply (case_tac "x:X")
apply (simp add: par_list_sync_any_any)
apply (simp add: par_list_interleave_any_any)
done

(*******************************
       par list head (if)
 *******************************)

(*** sync ***)

lemma par_list_if_sync: 
  "ALL X x u s t. x : X & u : s |[X]|list t
      --> x#u : (x#s) |[X]|list (x#t)"
apply (auto simp add: par_list_def)
done

(*** interleave left ***)

lemma par_list_if_left: 
  "ALL X x u s t. x ~: X & u : s |[X]|list t
      --> x#u : (x#s) |[X]|list t"
apply (intro allI)
apply (induct_tac t)
apply (auto simp add: par_list_def)
done

(*** interleave right ***)

lemma par_list_if_right: 
  "ALL X x u s t. x ~: X & u : s |[X]|list t
      --> x#u : s |[X]|list (x#t)"
apply (intro allI)
apply (induct_tac s)
apply (auto simp add: par_list_def)
done

(*** *** par list head (if) *** ***)

lemma par_list_head_if:
  "ALL X x u s t. 
     (x :X & (EX s'. s=x#s' & (EX t'. t=x#t' & u : s' |[X]|list t'))) |
     (x~:X & ((EX s'. s=x#s' & u : s' |[X]|list t ) |
              (EX t'. t=x#t' & u : s  |[X]|list t')))
        --> x#u : s |[X]|list t"
apply (intro allI impI)
apply (erule disjE)

apply (elim conjE exE)
apply (simp add: par_list_if_sync)

apply (erule conjE)
apply (erule disjE)

apply (erule exE)
apply (simp add: par_list_if_left)

apply (erule exE)
apply (simp add: par_list_if_right)
done

(*******************************
      par list head (iff)
 *******************************)

lemma par_list_head:
  "ALL X x u s t. x#u : s |[X]|list t
     = ((x :X & (EX s'. s=x#s' & (EX t'. t=x#t' & u : s' |[X]|list t'))) |
        (x~:X & ((EX s'. s=x#s' & u : s' |[X]|list t ) |
                 (EX t'. t=x#t' & u : s  |[X]|list t'))))"
apply (intro allI)
apply (rule iffI)
apply (simp add: par_list_head_only_if)
apply (simp add: par_list_head_if)
done

(*************************************************************
                    parallel list last
 *************************************************************)

(*******************************
    par list last (only if) 
 *******************************)

lemma par_list_last_only_if:
 "ALL X u x s t. u@[x] : s |[X]|list t
  --> ((x :X & (EX s'. s=s'@[x] & (EX t'. t=t'@[x] & u : s' |[X]|list t'))) |
       (x~:X & ((EX s'. s=s'@[x] & u : s' |[X]|list t ) |
                (EX t'. t=t'@[x] & u : s  |[X]|list t'))))"
apply (rule allI)
apply (rule allI)
apply (induct_tac u)

(* base case *)

apply (simp add: par_list_head par_list_nil)

(* step case *)

apply (intro allI impI)

apply (insert par_list_head_only_if)
apply (drule_tac x="X" in spec)
apply (rotate_tac -1)
apply (drule_tac x="a" in spec)
apply (drule_tac x="list @ [x]" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (erule disjE)

  (* sync *)

apply (elim conjE exE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (simp)


apply (erule disjE)

    (* sync sync *)

apply (elim conjE exE)
apply (simp add: par_list_head)

    (* sync left *)

apply (erule conjE)
apply (erule disjE)
apply (elim conjE exE)
apply (simp add: par_list_head)

    (* sync right *)

apply (elim conjE exE)
apply (simp add: par_list_head)

apply (erule conjE)
apply (erule disjE)

  (* left *)

apply (erule exE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t" in spec)

apply (simp)
apply (erule disjE)

    (* left sync *)

apply (elim conjE exE)
apply (simp add: par_list_head)

    (* left left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (simp add: par_list_head)

    (* left right *)

apply (elim conjE exE)
apply (simp add: par_list_head)

  (* right *)

apply (erule exE)
apply (drule_tac x="x" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t'" in spec)

apply (auto simp add: par_list_head)
done

(*******************************
       par list last (if)
 *******************************)

lemma par_list_last_if:
 "ALL X u x s t.
    ((x :X & (EX s'. s=s'@[x] & (EX t'. t=t'@[x] & u : s' |[X]|list t'))) |
     (x~:X & ((EX s'. s=s'@[x] & u : s' |[X]|list t ) |
              (EX t'. t=t'@[x] & u : s  |[X]|list t'))))
    --> u@[x] : s |[X]|list t"

apply (rule allI)
apply (rule allI)
apply (induct_tac u)

(* base case *)

apply (simp add: par_list_head par_list_nil)

(* step case : this can be done by (auto simp add: par_list_head) *)

apply (rename_tac X u y u')

apply (intro allI)
apply (intro impI)

apply (erule disjE)

  (* sync *)

apply (elim conjE exE)

apply (insert par_list_head_only_if)
apply (drule_tac x="X" in spec)
apply (rotate_tac -1)
apply (drule_tac x="y" in spec)
apply (drule_tac x="u'" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (simp)

apply (erule disjE)

    (* sync sync *)

apply (elim conjE exE)
apply (simp add: par_list_head)

    (* sync left *)

apply (erule conjE)
apply (erule disjE)
apply (elim conjE exE)
apply (simp add: par_list_head)

    (* sync right *)

apply (elim conjE exE)
apply (simp add: par_list_head)

apply (erule conjE)
apply (erule disjE)

  (* left *)

apply (erule exE)
apply (drule_tac x="X" in spec)
apply (rotate_tac -1)
apply (drule_tac x="y" in spec)
apply (drule_tac x="u'" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (erule disjE)

    (* left sync *)

apply (elim conjE exE)
apply (simp add: par_list_head)

    (* sync left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (simp add: par_list_head)

    (* sync right *)

apply (elim conjE exE)
apply (simp add: par_list_head)

  (* right *)

apply (erule exE)
apply (drule_tac x="X" in spec)
apply (rotate_tac -1)
apply (drule_tac x="y" in spec)
apply (drule_tac x="u'" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t'" in spec)
apply (simp)

apply (erule disjE)

    (* right sync *)

apply (elim conjE exE)
apply (simp add: par_list_head)

    (* right left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (simp add: par_list_head)

    (* right right *)

apply (elim conjE exE)
apply (simp add: par_list_head)
done

(*******************************
      par list last (iff)
 *******************************)

lemma par_list_last:
 "ALL X u x s t. u@[x] : s |[X]|list t
    = ((x :X & (EX s'. s=s'@[x] & (EX t'. t=t'@[x] & u : s' |[X]|list t'))) |
       (x~:X & ((EX s'. s=s'@[x] & u : s' |[X]|list t ) |
                (EX t'. t=t'@[x] & u : s  |[X]|list t'))))"
apply (intro allI)
apply (rule iffI)
apply (simp add: par_list_last_only_if)
apply (simp add: par_list_last_if)
done


(*************************************************************
                    symmetricity (list)
 *************************************************************)

lemma par_list_sym:
  "ALL X s t. (s |[X]|list t) = (t |[X]|list s)"
apply (rule allI)
apply (rule allI)
apply (induct_tac s)

      (* s=[] *)
apply (rule allI)
apply (induct_tac t)

          (* t=[] *)
apply (simp add: par_list_def)

          (* t~=[] *)
apply (simp add: par_list_def)

      (* s~=[] *)
apply (rule allI)
apply (induct_tac t)

          (* t=[] *)
apply (simp add: par_list_def)

          (* t=~[] *)

apply (rule equalityI)

           (* <= *)

apply (simp add: subset_def Ball_def)
apply (intro allI impI)

apply (insert list_nil_or_unnil)
apply (rotate_tac -1)
apply (drule_tac x="x" in spec)
apply (erule disjE)
apply (simp add: par_list_nil)
apply (erule exE)+
apply (simp add: par_list_head)

               (* ab : X |  ab ~: X  *)

apply (erule disjE)
apply (simp)
apply (erule conjE)
apply (erule disjE)
apply (simp)
apply (simp)

           (* <= *)

apply (simp add: subset_def Ball_def)
apply (intro allI impI)

apply (insert list_nil_or_unnil)
apply (rotate_tac -1)
apply (drule_tac x="x" in spec)
apply (erule disjE)
apply (simp add: par_list_nil)
apply (erule exE)+
apply (simp add: par_list_head)

               (* ab : X |  ab ~: X  *)

apply (erule disjE)
apply (simp)
apply (erule conjE)
apply (erule disjE)
apply (simp)
apply (simp)
done

(*************************************************************
         a new event is not introduced by |[X]|list
 *************************************************************)

(*** sync event ***)

lemma par_list_in_sync_event:
  "ALL X a u s t. a:X & u : s |[X]|list t & a : set u
      --> a : set s & a : set t"
apply (rule allI)
apply (rule allI)
apply (rule allI)

apply (induct_tac u)
apply (simp)

apply (intro allI impI)

apply (insert par_list_head_only_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="aa" in spec)
apply (rotate_tac -1)
apply (drule_tac x="list" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s" in spec)
apply (rotate_tac -1)
apply (drule_tac x="t" in spec)
apply (drule mp)
apply (erule conjE)+
apply (assumption)

  (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (simp)

apply (erule disjE)
apply (simp)
apply (simp)

  (* left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (erule disjE)
apply (simp)
apply (simp)

  (* right *)

apply (elim conjE exE)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t'" in spec)
apply (auto)
done

(*** interleave event ***)

lemma par_list_in_interleave_event:
  "ALL X a u s t. a~:X & u : s |[X]|list t & a : set u
      --> a : set s | a : set t"
apply (rule allI)
apply (rule allI)
apply (rule allI)

apply (induct_tac u)
apply (simp)

apply (intro allI)
apply (intro impI)

apply (insert par_list_head_only_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="aa" in spec)
apply (rotate_tac -1)
apply (drule_tac x="list" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s" in spec)
apply (rotate_tac -1)
apply (drule_tac x="t" in spec)
apply (simp)

  (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (simp)

apply (erule disjE)
apply (simp)
apply (simp)
apply (erule disjE)
apply (simp)
apply (simp)

  (* left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (erule disjE)
apply (simp)
apply (simp)

  (* right *)

apply (elim conjE exE)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t'" in spec)
apply (auto)
done

(*** sync event (contra pos) ***)

lemma contra_par_list_in_sync_event:
  "ALL X a s t. a:X & (a ~: set s | a ~: set t)
         --> (ALL u. u : s |[X]|list t --> a ~: set u)"
apply (intro allI impI)
apply (erule contrapos_pp)
apply (simp)

(* disj_not1 is deleted  --> apply (simp add: disj_not1) *)
(* disj_not1 is not deleted  --> apply (intro impI) *)
apply (simp add: disj_not1)

apply (insert par_list_in_sync_event)
apply (drule_tac x="X" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
done

(*** interleave event (contra pos) ***)

lemma contra_par_list_in_interleave_event:
  "ALL X a s t. a~:X & a ~: set s & a ~: set t
         --> (ALL u. u : s |[X]|list t --> a ~: set u)"
apply (intro allI impI)
apply (erule contrapos_pp)
apply (simp)

apply (case_tac "a:X")
apply (simp)

apply (insert par_list_in_interleave_event)
apply (drule_tac x="X" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
done

(*************************************************************
              every event remains in |[X]|list
 *************************************************************)

lemma par_list_in_sync_event_inv:
  "ALL X a u s t. a:X & u : s |[X]|list t & (a : set s | a : set t)
      --> a : set u"
apply (rule allI)
apply (rule allI)
apply (rule allI)

apply (induct_tac u)

  (* base case *)

apply (simp)
apply (rule disjI2)
apply (intro allI)
apply (simp add: disj_not1)
apply (intro allI impI)
apply (simp add: par_list_nil)

  (* step case *)

apply (intro allI impI)

apply (case_tac "aa = a")
apply (simp)

apply (simp)
apply (drule mp)

apply (simp add: par_list_head)

     (* sync *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t'" in exI)
apply (simp)
apply (blast)

     (* left *)

apply (erule conjE)
apply (rotate_tac -1)
apply (erule disjE)
apply (elim conjE exE)
apply (rule_tac x="s'" in exI)
apply (rule_tac x="t" in exI)
apply (simp)
apply (blast)

     (* right *)

apply (elim conjE exE)
apply (rule_tac x="s" in exI)
apply (rule_tac x="t'" in exI)
apply (simp)
apply (blast)

apply (simp)
done

(*** sync event (contra pos) ***)

lemma contra_par_list_in_sync_event_inv:
  "ALL X a u s t. a:X & u : s |[X]|list t & a ~: set u
         --> a ~: set s & a ~: set t"
apply (intro allI impI)
apply (erule contrapos_pp)
apply (simp)

apply (simp add: disj_not1)
apply (intro impI)

apply (insert par_list_in_sync_event_inv)
apply (drule_tac x="X" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
done


(*************************************************************
                    parallel list app type
 *************************************************************)

lemma par_list_in_trace_only_if:
 "ALL X v u s t. v : trace & u : trace & s : trace & t : trace &
      u : s |[X]|list t & v @ u : trace --> v @ s : trace & v @ t : trace"
apply (intro allI impI)
apply (erule conjE)+
apply (simp add: list_app_in_trace)
apply (erule disjE)
apply (simp)
apply (simp add: par_list_nil)
done

lemma par_list_in_trace_if:
 "ALL X v u s t. v : trace & u : trace & s : trace & t : trace &
      u : s |[X]|list t &  v @ s : trace & v @ t : trace --> v @ u : trace"
apply (intro allI impI)
apply (erule conjE)+
apply (simp add: list_app_in_trace)
apply (erule disjE)
apply (simp)
apply (erule disjE)
apply (simp)
apply (simp add: par_list_def)
done

(*************************************************************
                 parallel list lemmas etc.
 *************************************************************)

(*******************************
         par list lenght
 *******************************)

lemma par_list_length:
 "ALL X u s t. u : s |[X]|list t
         --> length s <= length u & length t <= length u"
apply (rule allI)
apply (rule allI)
apply (induct_tac u)
apply (simp add: par_list_nil)

apply (intro allI impI)

apply (insert par_list_head_only_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="a" in spec)
apply (rotate_tac -1)
apply (drule_tac x="list" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s" in spec)
apply (rotate_tac -1)
apply (drule_tac x="t" in spec)
apply (simp)

  (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (simp)

  (* left *)

apply (erule conjE)
apply (erule disjE)
apply (elim conjE exE)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t" in spec)
apply (auto)

  (* right *)

apply (drule_tac x="s" in spec)
apply (drule_tac x="t'" in spec)
apply (auto)
done

(*************************************************************
                  parallel rep in trace
 *************************************************************)

(*** par list in trace ***)

lemma par_list_in_trace:
 "ALL X (s::'a event list) t u.
     s: trace & t: trace & u : (s |[X Un {Tick}]|list t) --> u: trace"
apply (intro allI impI)
apply (elim conjE)

apply (insert list_last_nil_or_unnil)
apply (drule_tac x="u" in spec)
apply (erule disjE)

apply (simp add: trace_def)

apply (erule exE)+
apply (simp (no_asm) add: trace_def)
apply (simp)

apply (simp add: par_list_last)

  (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (simp add: trace_def)
apply (insert contra_par_list_in_sync_event)
apply (drule_tac x="insert Tick X" in spec)
apply (drule_tac x="Tick" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (drule mp)
apply (simp)
apply (drule_tac x="sa" in spec)
apply (simp)

  (* left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (simp add: trace_def)

apply (drule_tac x="insert Tick X" in spec)
apply (drule_tac x="Tick" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t" in spec)
apply (drule mp)
apply (simp)
apply (drule_tac x="sa" in spec)
apply (simp)

  (* right *)

apply (unfold trace_def)
apply (elim conjE exE)
apply (drule_tac x="insert Tick X" in spec)
apply (drule_tac x="Tick" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t'" in spec)
apply (drule mp)
apply (simp)
apply (drule_tac x="sa" in spec)
apply (simp)
done

(***  par rep in trace ***)

lemma par_rep_in_trace:
 "ALL X (s::'a trace) t u. u : (s |[X]|rep t) --> u: trace"
apply (intro allI impI)
apply (simp add: par_rep_def)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="u" in spec)

apply (simp add: Rep_trace)
done

(*******************************
        par list in trace
 *******************************)

lemma par_list_set_in_trace:
  "ALL X s t. s: trace & t: trace 
       --> (ALL u : s |[X Un {Tick}]|list t. u: trace)"
apply (intro allI impI)
apply (simp add: Ball_def)
apply (intro allI impI)
apply (insert par_list_in_trace)
apply (drule_tac x="X" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (drule_tac x="x" in spec)
apply (simp)
done

(*************************************************************
                 parallel list prefix_closed
 *************************************************************)

lemma par_list_prefix:
 "ALL X v u s t. s: trace & t: trace & prefix_list v u & u : s |[X]|list t 
     --> (EX s' t'. v : s' |[X]|list t' & 
                        prefix_list s' s & prefix_list t' t)"
apply (rule allI)
apply (rule allI)
apply (induct_tac v)

 (* base case *)

apply (intro allI impI)
apply (rule_tac x="[]" in exI)
apply (rule_tac x="[]" in exI)
apply (simp add: par_list_def prefix_list_def)

 (* step case *)

apply (intro allI impI)

apply (insert prefix_list_head_same)
apply (drule_tac x="a" in spec)
apply (rotate_tac -1)
apply (drule_tac x="list" in spec)
apply (rotate_tac -1)
apply (drule_tac x="u" in spec)
apply (simp)
apply (erule exE)

apply (simp add: par_list_head)
apply (erule conjE)+

  (* a : X *)

apply (erule disjE)
apply (elim conjE exE)

apply (drule_tac x="u'" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)

apply (insert event_app_in_trace_inv)
apply (frule_tac x="a" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (simp)

apply (erule exE)+
apply (rename_tac X a list u s t u' s' t' s'' t'')

apply (rule_tac x="a # s''" in exI)
apply (rule_tac x="a # t''" in exI)
apply (simp add: prefix_list_def)

  (* (EX s'. s = a # s' & u' : s' |[X]|list t) *)

apply (erule conjE)
apply (erule disjE)
apply (elim conjE exE)

apply (drule_tac x="u'" in spec)
 apply (drule_tac x="a" in spec)
apply (drule_tac x="s'" in spec)
 apply (drule_tac x="s'" in spec)
apply (drule_tac x="t" in spec)
apply (simp)

apply (erule exE)+
apply (rename_tac X a list u s t u' s' s'' t')
apply (rule_tac x="a # s''" in exI)
apply (rule_tac x="t'" in exI)
apply (simp add: prefix_list_def)

  (* (EX t'. t = a # t' & u' : s |[X]|list t') *)

apply (elim conjE exE)

apply (drule_tac x="u'" in spec)
 apply (drule_tac x="a" in spec)
apply (drule_tac x="s" in spec)
 apply (drule_tac x="t'" in spec)
apply (drule_tac x="t'" in spec)
apply (simp)

apply (erule exE)+
apply (rename_tac X a list u s t u' t' s' t'')
apply (rule_tac x="s'" in exI)
apply (rule_tac x="a # t''" in exI)
apply (simp add: prefix_list_def)
done

(*************************************************************
         a new event is not introduced by |[X]|set
 *************************************************************)

(*** sync event (contra pos) ***)

lemma contra_par_set_in_sync_event:
  "ALL X a s t. a: (Ev ` X) Un {Tick} & (a ~: sett s | a ~: sett t)
         --> (ALL u. u : s |[X]|set t --> a ~: sett u)"
apply (intro allI impI)
apply (simp add: par_rep_def)

apply (insert contra_par_list_in_sync_event)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule mp)
apply (simp)
apply (drule_tac x="Rep_trace u" in spec)

apply (drule mp)

apply (insert t_in_Abs_set)
apply (drule_tac x="u" in spec)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)
apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="x" in spec)
apply (simp add: Rep_trace)
apply (simp)
apply (simp)
done

(*************************************************************
              every event remains in |[X]|set
 *************************************************************)

lemma contra_par_set_in_sync_event_inv:
  "ALL X a u s t. a: (Ev ` X) Un {Tick} & u : s |[X]|set t & a ~: sett u
       --> a ~: sett s & a ~: sett t"
apply (intro allI impI)
apply (simp add: par_rep_def)

apply (insert contra_par_list_in_sync_event_inv)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="a" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule mp)
apply (simp)

apply (insert t_in_Abs_set)
apply (drule_tac x="u" in spec)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)
apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="x" in spec)
apply (simp add: Rep_trace)
apply (simp)
apply (simp)
done

(*************************************************************
                  parallel trace app type
 *************************************************************)

lemma par_trace_in_trace_if:
 "ALL X v u s t. u : s |[X]|set t &  v @rep s : trace & v @rep t : trace 
         --> v @rep u : trace"
apply (intro allI impI)

apply (insert par_list_in_trace_if)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="Rep_trace v" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (simp add: Rep_trace)
apply (drule mp)

apply (simp add: par_rep_def)
apply (insert t_in_Abs_set)
apply (drule_tac x="u" in spec)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)
apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="x" in spec)
apply (simp add: Rep_trace)
apply (simp)
apply (simp)
done

(*===========================================================*
 |                                                           |
 |                    list --> trace                         |
 |                                                           |
 *===========================================================*)

(*************************************************************
                    parallel trace nil
 *************************************************************)

lemma par_trace_nil:
  "ALL X s t. ([]t : s |[X]|set t) = (s = []t & t = []t)"
apply (intro allI)
apply (simp add: par_rep_def)
apply (insert par_list_nil)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (rule iffI)

      (* only if *)

apply (insert t_in_Abs_set)
apply (drule_tac x="[]t" in spec)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t" 
       in spec)
apply (simp)
apply (drule mp)

apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="x" in spec)
apply (simp add: Rep_trace)
apply (simp add: nil_in_trace Abs_trace_inverse Abs_Rep_t_inv_imp)

      (* if *)

apply (simp add: Abs_Rep_t_inv nil_in_trace)
done

(*************************************************************
                    parallel trace head
 *************************************************************)

(*******************************
    par trace head (only if)
 *******************************)

lemma par_trace_head_only_if:
 "ALL X x u s t. [x]t @rep u : trace & [x]t @t u : s |[X]|set t
    --> ((x : Ev ` X Un {Tick}
               & (EX s' t'. u : s' |[X]|set t' & 
                            s = [x]t @t s' & [x]t @rep s': trace &
                            t = [x]t @t t' & [x]t @rep t': trace)) |
         (x~: Ev ` X Un {Tick}
               & ((EX s'. u : s' |[X]|set t  & 
                            s = [x]t @t s' & [x]t @rep s': trace) |
                  (EX t'. u : s  |[X]|set t' & 
                            t = [x]t @t t' & [x]t @rep t': trace))))"
apply (intro allI impI)
apply (insert par_list_head_only_if)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule mp)

  (* trace assumption --> list assumption *)

apply (simp add: par_rep_def)
apply (insert t_in_Abs_set)
apply (drule_tac x="[x]t @t u" in spec)
apply (drule_tac x="(Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t)" 
       in spec)
apply (drule mp)
apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="xa" in spec)
apply (simp add: Rep_trace)

apply (simp add: event_in_trace Abs_trace_inverse)

  (* list conclusion --> trace conclusion *)

apply (simp add: par_rep_def)

    (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI1)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (insert event_app_in_trace_inv)
apply (frule_tac x="x" in spec)
apply (drule_tac x="x" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s'" in spec)
apply (rotate_tac -2)
apply (drule_tac x="t'" in spec)
apply (simp add: Rep_trace_eq)

apply (rule_tac x="Abs_trace s'" in exI)
apply (rule_tac x="Abs_trace t'" in exI)
apply (simp add: Abs_Rep_t_inv_imp Abs_trace_inverse)
apply (simp add: t_in_Rep_set)

apply (insert par_list_in_trace_only_if)
apply (rotate_tac -1)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="[x]" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (simp add: event_in_trace Rep_trace)

    (* left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI2)
apply (simp)

apply (rule disjI1)
apply (simp add: event_in_trace Abs_trace_inverse)

      (* event_app_in_trace_inv *)
apply (drule_tac x="x" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s'" in spec)
apply (simp add: Rep_trace_eq)

apply (rule_tac x="Abs_trace s'" in exI)
apply (simp add: Abs_Rep_t_inv_imp Abs_trace_inverse)
apply (simp add: t_in_Rep_set)

      (* insert par_list_in_trace_only_if *)
apply (rotate_tac 2)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="[x]" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="s'" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (simp add: event_in_trace Rep_trace)

    (* right *)

apply (elim conjE exE)
apply (rule disjI2)
apply (rule conjI)
apply (assumption)
apply (rule conjI)
apply (assumption)

apply (rule disjI2)

      (* event_app_in_trace_inv *)
apply (drule_tac x="x" in spec)
apply (rotate_tac -1)
apply (drule_tac x="t'" in spec)

apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: Rep_trace_eq)

apply (rule_tac x="Abs_trace t'" in exI)
apply (simp add: Abs_Rep_t_inv_imp Abs_trace_inverse)
apply (simp add: t_in_Rep_set)

      (* insert par_list_in_trace_only_if *)
apply (rotate_tac 2)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="[x]" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="t'" in spec)
apply (simp add: event_in_trace Rep_trace)
done

(*******************************
      par trace head (if)
 *******************************)

lemma par_trace_head_if:
 "ALL X x u s t.
         ((x : Ev ` X Un {Tick}
               & (EX s' t'. u : s' |[X]|set t' & 
                            s = [x]t @t s' & [x]t @rep s': trace &
                            t = [x]t @t t' & [x]t @rep t': trace)) |
         (x~: Ev ` X Un {Tick}
               & ((EX s'. u : s' |[X]|set t  & 
                            s = [x]t @t s' & [x]t @rep s': trace) |
                  (EX t'. u : s  |[X]|set t' & 
                            t = [x]t @t t' & [x]t @rep t': trace))))
    --> [x]t @rep u : trace & [x]t @t u : s |[X]|set t"
apply (intro allI impI)
apply (insert par_list_head_if)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule mp)

  (* trace assumption --> list assumption *)

apply (simp add: par_rep_def)
(* apply (erule conjE)+ *)

    (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI1)
apply (simp)

apply (simp add: rep_appt_to_apprep)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (insert t_in_Abs_set)
apply (drule_tac x="u" in spec)
apply (drule_tac x="Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t'"
       in spec)
apply (drule mp)

        (* Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t' TYPE? *)

apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s'" in spec)
apply (drule_tac x="Rep_trace t'" in spec)
apply (drule_tac x="xa" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
apply (simp)

    (* left *)

apply (simp)
apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI1)

apply (simp add: rep_appt_to_apprep)
apply (simp add: event_in_trace Abs_trace_inverse)

      (* t_in_Abs_set *)
apply (drule_tac x="u" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)

        (* Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t TYPE? *)

apply (simp add: Ball_def)
apply (intro allI impI)

      (* insert par_list_in_trace *)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s'" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="xa" in spec)
apply (simp add: Rep_trace)
apply (assumption)

    (* right *)

apply (elim conjE exE)
apply (rule disjI2)

      (* t_in_Abs_set *)
apply (drule_tac x="u" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t'"
       in spec)
apply (drule mp)

        (* Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t TYPE? *)

apply (unfold Ball_def)
apply (rule conjI)
apply (intro allI impI)

      (* insert par_list_in_trace *)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t'" in spec)
apply (drule_tac x="xa" in spec)
apply (simp add: Rep_trace)
apply (assumption)

apply (drule_tac x="{}" in spec) (* dummy *)
apply (drule_tac x="[]" in spec) (* dummy *)
apply (drule_tac x="[]" in spec) (* dummy *)
apply (drule_tac x="[]" in spec) (* dummy *)

apply (simp add: rep_appt_to_apprep)
apply (simp add: event_in_trace Abs_trace_inverse)

  (* list conclusion --> trace conclusion *)

       (* sync *)

apply (erule disjE)
apply (elim conjE exE)

          (* for [x]t @rep u : trace u *)

apply (insert par_trace_in_trace_if)
apply (rotate_tac -1)
apply (drule_tac x="X" in spec)
apply (rotate_tac -1)
apply (drule_tac x="[x]t" in spec)
apply (rotate_tac -1)
apply (drule_tac x="u" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s'" in spec)
apply (rotate_tac -1)
apply (drule_tac x="t'" in spec)
apply (simp)

apply (insert t_in_Rep_set)
apply (rotate_tac -1)
apply (drule_tac x="[x]t @t u" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (simp add: par_rep_def)

       (* x ~= Tick (left and right) *)

apply (erule conjE)

          (* for [x]t @rep u : trace u *)

apply (insert notick_app_in_trace)
apply (drule_tac x="x" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace u" in spec)
apply (simp add: par_rep_def Rep_trace event_in_trace Abs_trace_inverse)
done

(*** this E-rule is used to remove this from assumptions ***)

lemma par_trace_head_if_remE:
 "[| ALL X x u s t.
         ((x : Ev ` X Un {Tick}
               & (EX s' t'. u : s' |[X]|set t' & 
                            s = [x]t @t s' & [x]t @rep s': trace &
                            t = [x]t @t t' & [x]t @rep t': trace)) |
         (x~: Ev ` X Un {Tick}
               & ((EX s'. u : s' |[X]|set t  & 
                            s = [x]t @t s' & [x]t @rep s': trace) |
                  (EX t'. u : s  |[X]|set t' & 
                            t = [x]t @t t' & [x]t @rep t': trace))))
    --> [x]t @rep u : trace & [x]t @t u : s |[X]|set t;
   [| True |] ==> R |] ==> R"
apply (insert par_trace_head_if)
apply (simp)
done

(*******************************
     par trace head (iff)
 *******************************)

lemma par_trace_head:
 "ALL X x u s t. ([x]t @rep u : trace & [x]t @t u : s |[X]|set t)
    = ((x : Ev ` X Un {Tick}
               & (EX s' t'. u : s' |[X]|set t' & 
                            s = [x]t @t s' & [x]t @rep s': trace &
                            t = [x]t @t t' & [x]t @rep t': trace)) |
         (x~: Ev ` X Un {Tick}
               & ((EX s'. u : s' |[X]|set t  & 
                            s = [x]t @t s' & [x]t @rep s': trace) |
                  (EX t'. u : s  |[X]|set t' & 
                            t = [x]t @t t' & [x]t @rep t': trace))))"
apply (intro allI)
apply (rule iffI)
apply (simp add: par_trace_head_only_if 
            del: Un_insert_right Un_iff conj_assoc)
apply (insert par_trace_head_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
done

(*************************************************************
                    parallel trace last
 *************************************************************)

(*******************************
    par trace last (only if)
 *******************************)

lemma par_trace_last_only_if:
 "ALL X u x s t. notick u & u @t [x]t : s |[X]|set t
    --> ((x : Ev ` X Un {Tick}
               & (EX s' t'. u : s' |[X]|set t' & 
                            s = s' @t [x]t & notick s' &
                            t = t' @t [x]t & notick t')) |
         (x~: Ev ` X Un {Tick}
               & ((EX s'. u : s' |[X]|set t  &
                            s = s' @t [x]t & notick s') |
                  (EX t'. u : s  |[X]|set t' &
                            t = t' @t [x]t & notick t'))))"
apply (intro allI impI)
apply (insert par_list_last_only_if)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule mp)

  (* trace assumption --> list assumption *)

apply (simp add: par_rep_def)
apply (insert t_in_Abs_set)
apply (drule_tac x="u @t [x]t" in spec)
apply (drule_tac x="(Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t)" 
       in spec)
apply (drule mp)
apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="xa" in spec)
apply (simp)
apply (simp add: Rep_trace)

apply (insert trace_app_in_trace_if)
apply (drule_tac x="u" in spec)
apply (drule_tac x="[x]t" in spec)

apply (simp add: event_in_trace Abs_trace_inverse)

  (* list conclusion --> trace conclusion *)

apply (simp add: par_rep_def)

    (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI1)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (insert app_event_notick_inv)
apply (rotate_tac -1)
apply (frule_tac x="s'" in spec)
apply (drule_tac x="t'" in spec)
apply (rotate_tac -1)
apply (drule_tac x="x" in spec)
apply (rotate_tac -2)
apply (drule_tac x="x" in spec)
apply (simp add: Rep_trace_eq)

apply (rule_tac x="Abs_trace s'" in exI)
apply (rule_tac x="Abs_trace t'" in exI)
apply (simp add: Abs_Rep_t_inv_imp Abs_trace_inverse)
apply (simp add: t_in_Rep_set)

    (* left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI2)
apply (simp)

apply (rule disjI1)
apply (simp add: event_in_trace Abs_trace_inverse)

      (* app_event_notick_inv *)
apply (rotate_tac 3)
apply (drule_tac x="s'" in spec)
apply (drule mp)
apply (rule_tac x="x" in exI)
apply (simp add: Rep_trace_eq)

apply (rule_tac x="Abs_trace s'" in exI)
apply (simp add: Abs_Rep_t_inv_imp Abs_trace_inverse)
apply (simp add: t_in_Rep_set)

    (* right *)

apply (elim conjE exE)
apply (rule disjI2)
apply (rule conjI)
apply (assumption)
apply (rule conjI)
apply (assumption)

apply (rule disjI2)

      (* app_event_notick_inv *)
apply (rotate_tac 3)
apply (drule_tac x="t'" in spec)
apply (drule_tac x="x" in spec)

apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: Rep_trace_eq)

apply (rule_tac x="Abs_trace t'" in exI)
apply (simp add: Abs_Rep_t_inv_imp Abs_trace_inverse)
apply (simp add: t_in_Rep_set)
done

(*******************************
       par trace last (if)
 *******************************)

lemma par_trace_last_if:
 "ALL X u x s t. 
         ((x : Ev ` X Un {Tick}
               & (EX s' t'. u : s' |[X]|set t' & 
                            s = s' @t [x]t & notick s' &
                            t = t' @t [x]t & notick t')) |
         (x~: Ev ` X Un {Tick}
               & ((EX s'. u : s' |[X]|set t  &
                            s = s' @t [x]t & notick s') |
                  (EX t'. u : s  |[X]|set t' &
                            t = t' @t [x]t & notick t'))))
         --> u @t [x]t : s |[X]|set t & notick u"
apply (intro allI impI)
apply (insert par_list_last_if)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule mp)

  (* trace assumption --> list assumption *)

apply (simp add: par_rep_def)
(* apply (erule conjE)+ *)

    (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI1)
apply (simp)

apply (simp add: rep_appt_to_apprep)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (insert t_in_Abs_set)
apply (drule_tac x="u" in spec)
apply (drule_tac x="Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t'"
       in spec)
apply (drule mp)

        (* Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t' TYPE? *)

apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s'" in spec)
apply (drule_tac x="Rep_trace t'" in spec)
apply (drule_tac x="xa" in spec)
apply (simp add: Rep_trace Rep_trace_inverse)
apply (simp)

    (* left *)

apply (simp)
apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)
apply (rule disjI1)

apply (simp add: rep_appt_to_apprep)
apply (simp add: event_in_trace Abs_trace_inverse)

      (* t_in_Abs_set *)
apply (drule_tac x="u" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)

        (* Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t TYPE? *)

apply (simp add: Ball_def)
apply (intro allI impI)

      (* insert par_list_in_trace *)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s'" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="xa" in spec)
apply (simp add: Rep_trace)
apply (simp)

    (* right *)

apply (elim conjE exE)
apply (rule disjI2)

      (* t_in_Abs_set *)
apply (drule_tac x="u" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t'"
       in spec)
apply (drule mp)

        (* Rep_trace s' |[insert Tick (Ev ` X)]|list Rep_trace t TYPE? *)

apply (unfold Ball_def)
apply (rule conjI)
apply (intro allI impI)

      (* insert par_list_in_trace *)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t'" in spec)
apply (drule_tac x="xa" in spec)
apply (simp add: Rep_trace)
apply (assumption)

apply (drule_tac x="{}" in spec) (* dummy *)
apply (drule_tac x="[]" in spec) (* dummy *)
apply (drule_tac x="[]" in spec) (* dummy *)
apply (drule_tac x="[]" in spec) (* dummy *)

apply (simp add: rep_appt_to_apprep)
apply (simp add: event_in_trace Abs_trace_inverse)

  (* list conclusion --> trace conclusion *)

       (* sync *)

apply (erule disjE)
apply (elim conjE exE)

          (* for notick u *)

apply (insert contra_par_set_in_sync_event)
apply (rotate_tac -1)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Tick" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s'" in spec)
apply (rotate_tac -1)
apply (drule_tac x="t'" in spec)
apply (drule mp)
apply (simp)
apply (rotate_tac -1)
apply (drule_tac x="u" in spec)
apply (simp)

apply (insert t_in_Rep_set)
apply (rotate_tac -1)
apply (drule_tac x="u @t [x]t" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)
apply (simp add: notick_app_trace_rep)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (simp add: par_rep_def)

       (* left *)

apply (erule conjE)+
apply (erule disjE)
apply (elim conjE exE)

          (* for notick u *)

      (* contra_par_set_in_sync_event *)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Tick" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s'" in spec)
apply (rotate_tac -1)
apply (drule_tac x="t" in spec)
apply (drule mp)
apply (simp)
apply (rotate_tac -1)
apply (drule_tac x="u" in spec)
apply (simp)

      (* insert t_in_Rep_set *)
apply (rotate_tac 3)
apply (drule_tac x="u @t [x]t" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)
apply (simp add: notick_app_trace_rep)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (simp add: par_rep_def)

       (* right *)

apply (elim conjE exE)

          (* for notick u *)

      (* contra_par_set_in_sync_event *)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Tick" in spec)
apply (rotate_tac -1)
apply (drule_tac x="s" in spec)
apply (rotate_tac -1)
apply (drule_tac x="t'" in spec)
apply (drule mp)
apply (simp)
apply (rotate_tac -1)
apply (drule_tac x="u" in spec)
apply (simp)

      (* t_in_Rep_set *)
apply (rotate_tac 3)
apply (drule_tac x="u @t [x]t" in spec)
apply (rotate_tac -1)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)
apply (simp add: notick_app_trace_rep)
apply (simp add: event_in_trace Abs_trace_inverse)

apply (simp add: par_rep_def)
done

(*******************************
     par trace last (iff)
 *******************************)

lemma par_trace_last:
 "ALL X u x s t. (notick u & u @t [x]t : s |[X]|set t)
    = ((x : Ev ` X Un {Tick}
               & (EX s' t'. u : s' |[X]|set t' & 
                            s = s' @t [x]t & notick s' &
                            t = t' @t [x]t & notick t')) |
         (x~: Ev ` X Un {Tick}
               & ((EX s'. u : s' |[X]|set t  &
                            s = s' @t [x]t & notick s') |
                  (EX t'. u : s  |[X]|set t' &
                            t = t' @t [x]t & notick t'))))"
apply (intro allI)
apply (rule iffI)
apply (simp add: par_trace_last_only_if 
            del: Un_insert_right Un_iff conj_assoc)
apply (insert par_trace_last_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="u" in spec)
apply (drule_tac x="x" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (simp)
done

(*************************************************************
                    parallel trace nil
 *************************************************************)

(*******************************
    par trace nil nil
 *******************************)

lemma par_trace_nil_nil:
 "ALL X. ([]t : []t |[X]|set []t)"
apply (intro allI)
apply (simp add: par_rep_def)
apply (simp add: nil_in_trace Abs_trace_inverse)
apply (simp add: par_list_def)
done

(*******************************
    par trace nil nil -> nil
 *******************************)

lemma par_trace_nil_nil_to_nil:
 "ALL X s. (s : []t |[X]|set []t) --> s = []t"
apply (intro allI impI)
apply (simp add: par_rep_def)
apply (simp add: nil_in_trace Abs_trace_inverse)
apply (simp add: par_list_def)
done

(*************************************************************
                    parallel trace Tick
 *************************************************************)

(*******************************
    par trace Tick (only if)
 *******************************)

lemma par_trace_Tick_only_if:
 "ALL X s t. ([Tick]t : s |[X]|set t) --> (s = [Tick]t & t = [Tick]t)"
apply (intro allI impI)
apply (insert par_trace_head_only_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Tick" in spec)
apply (drule_tac x="[]t" in spec)
apply (drule_tac x="s" in spec)
apply (drule_tac x="t" in spec)
apply (drule mp)

apply (simp add: nil_in_trace event_in_trace Abs_trace_inverse)

      (* sync *)

apply (erule disjE)
apply (elim conjE exE)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (insert Tick_app_nil_in_trace)
apply (frule_tac x="Rep_trace s'" in spec)
apply (drule_tac x="Rep_trace t'" in spec)
apply (simp)

      (* no sync --> contradiction *)

apply (simp)
done

(*******************************
       par trace Tick (if)
 *******************************)

lemma par_trace_Tick_if: "[Tick]t : [Tick]t |[X]|set [Tick]t"
apply (insert par_trace_head_if)
apply (drule_tac x="X" in spec)
apply (drule_tac x="Tick" in spec)
apply (drule_tac x="[]t" in spec)
apply (drule_tac x="[Tick]t" in spec)
apply (drule_tac x="[Tick]t" in spec)
apply (drule mp)

apply (simp)
apply (rule_tac x="[]t" in exI)
apply (rule_tac x="[]t" in exI)
apply (simp add: par_trace_nil)
apply (simp add: nil_in_trace event_in_trace Abs_trace_inverse)

apply (simp add: nil_in_trace event_in_trace Abs_trace_inverse)
done

(*******************************
     par trace Tick (iff)
 *******************************)

lemma par_trace_Tick:
 "ALL X s t. ([Tick]t : s |[X]|set t) = (s = [Tick]t & t = [Tick]t)"
apply (intro allI)
apply (rule iffI)
apply (simp add: par_trace_Tick_only_if)
apply (simp add: par_trace_Tick_if)
done

(**********************************
     par trace Tick Tick -> Tick
 **********************************)

lemma par_trace_Tick_Tick_to_Tick:
 "ALL X s. (s : [Tick]t |[X]|set [Tick]t) --> s = [Tick]t"
apply (intro allI impI)
apply (simp add: par_rep_def)
apply (simp add: event_in_trace Abs_trace_inverse)
apply (simp add: par_list_def)
done

(*************************************************************
                    symmetricity (trace)
 *************************************************************)

lemma par_trace_sym:
  "ALL X s t. (s |[X]|set t) = (t |[X]|set s)"
apply (intro allI)
apply (simp add: par_rep_def)
apply (simp add: par_list_sym)
done

(*************************************************************
                 parallel trace prefix_closed
 *************************************************************)

lemma par_trace_prefix:
 "ALL X v u s t. prefix v u & u : s |[X]|set t 
     --> (EX s' t'. v : s' |[X]|set t' & 
                        prefix s' s & prefix t' t)"
apply (intro allI impI)
apply (insert par_list_prefix)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="Rep_trace v" in spec)
apply (drule_tac x="Rep_trace u" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule mp)

  (* trace assumption --> list assumption *)

apply (simp add: par_rep_def)
apply (insert t_in_Abs_set)
apply (drule_tac x="u" in spec)
apply (drule_tac x="Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t"
       in spec)
apply (drule mp)
apply (simp add: Ball_def)
apply (intro allI impI)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="x" in spec)
apply (simp add: Rep_trace)
apply (simp add: Rep_trace)


  (* list conclusion --> trace conclusion *)

apply (erule exE)+
apply (rule_tac x="Abs_trace s'" in exI)
apply (rule_tac x="Abs_trace t'" in exI)

apply (simp add: par_rep_def)
apply (erule conjE)+
apply (simp add: prefix_list_s_trace Abs_trace_inverse)
apply (simp add: t_in_Rep_set)
done

(*************************************************************
                 parallel trace lemmas etc.
 *************************************************************)

(*******************************
        par trace lenght 
 *******************************)

lemma par_trace_length:
 "ALL X u s t. u : s |[X]|set t
         --> lengtht s <= lengtht u & lengtht t <= lengtht u"
apply (intro allI impI)
apply (simp add: par_rep_def)
apply (insert in_image_set)
apply (drule_tac x="u" in spec)
apply (drule_tac x="Abs_trace" in spec)
apply (drule_tac x=
   "Rep_trace s |[insert Tick (Ev ` X)]|list Rep_trace t" in spec)
apply (simp)
apply (elim conjE exE)

apply (insert par_list_length)
apply (drule_tac x="insert Tick (Ev ` X)" in spec)
apply (drule_tac x="y" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (simp)

apply (insert par_list_in_trace)
apply (drule_tac x="Ev ` X" in spec)
apply (drule_tac x="Rep_trace s" in spec)
apply (drule_tac x="Rep_trace t" in spec)
apply (drule_tac x="y" in spec)
apply (simp add: Rep_trace Abs_trace_inverse)
done

(**************************************************
            u : []t |[X]|set ([Ev a]t @t t)
 **************************************************)

lemma par_trace_head_nil_Ev_app:
  "ALL X u a t. u : []t |[X]|set ([Ev a]t @t t)
     --> a ~: X & (EX v. u = [Ev a]t @t v & v : []t |[X]|set t)"
apply (intro allI impI)
apply (insert trace_nil_or_Tick_or_Ev)
apply (drule_tac x="u" in spec)

 (* u=[]t --> contra *)
 apply (erule disjE)
 apply (simp add: par_trace_nil)
 apply (simp add: Abs_trace_inject nil_in_trace trace_event_app_in_trace)
 apply (simp add: Abs_trace_inverse event_in_trace)

 (* u=[Tick]t --> contra *)
 apply (erule disjE)
 apply (simp add: par_trace_Tick)
 apply (simp add: Abs_trace_inject nil_in_trace event_in_trace)

 (* x=[Ev a]t @t s *)
 apply (elim exE)
 apply (insert par_trace_head_only_if)

 apply (drule_tac x="X" in spec)
 apply (drule_tac x="Ev aa" in spec)
 apply (drule_tac x="s" in spec)
 apply (drule_tac x="[]t" in spec)
 apply (drule_tac x="([Ev a]t @t t)" in spec)
 apply (simp add: trace_event_app_in_trace)

 apply (erule disjE)

  (* Ev aa : Ev ` X --> contra *)
  apply (elim conjE exE)
  apply (rotate_tac -2)
  apply (drule sym)
  apply (simp add: notick_Ev app_t_nil)
  apply (simp add: nil_in_trace event_in_trace Abs_trace_inject)

  (* Ev aa ~: Ev ` X *)
  apply (elim conjE)
  apply (erule disjE)

   (* left step --> contra *)
   apply (elim conjE exE)
   apply (rotate_tac -1)
   apply (drule sym)
   apply (simp add: notick_Ev app_t_nil)
   apply (simp add: nil_in_trace event_in_trace Abs_trace_inject)

   (* right step *)
   apply (elim conjE exE)
   apply (insert same_head_tail)
   apply (drule_tac x="a" in spec)
   apply (drule_tac x="aa" in spec)
   apply (drule_tac x="t" in spec)
   apply (drule_tac x="t'" in spec)

   apply (simp add: Rep_trace_inject image_iff)
   apply (rule_tac x="s" in exI)
   apply (simp)
done

(**************************************************
          u : [Tick]t |[X]|set ([Ev a]t @t t)
 **************************************************)

lemma par_trace_head_Tick_Ev_app:
  "ALL X u a t. u : [Tick]t |[X]|set ([Ev a]t @t t)
     --> a ~: X & (EX v. u = [Ev a]t @t v & v : [Tick]t |[X]|set t)"
apply (intro allI impI)
apply (insert trace_nil_or_Tick_or_Ev)
apply (drule_tac x="u" in spec)

 (* u=[]t --> contra *)
 apply (erule disjE)
 apply (simp add: par_trace_nil)
 apply (simp add: Abs_trace_inject nil_in_trace trace_event_app_in_trace)
 apply (simp add: Abs_trace_inverse event_in_trace)

 (* u=[Tick]t --> contra *)
 apply (erule disjE)
 apply (simp add: par_trace_Tick)
 apply (simp add: Abs_trace_inject trace_event_app_in_trace event_in_trace)
 apply (simp add: event_in_trace Abs_trace_inverse)

 (* x=[Ev a]t @t s *)
 apply (elim exE)
 apply (insert par_trace_head_only_if)
 apply (drule_tac x="X" in spec)
 apply (drule_tac x="Ev aa" in spec)
 apply (drule_tac x="s" in spec)
 apply (drule_tac x="[Tick]t" in spec)
 apply (drule_tac x="([Ev a]t @t t)" in spec)
 apply (simp add: trace_event_app_in_trace)

 apply (erule disjE)

  (* Ev aa : Ev ` X --> contra *)
  apply (elim conjE exE)
  apply (simp add: Abs_trace_inject trace_event_app_in_trace event_in_trace)
  apply (simp add: event_in_trace Abs_trace_inverse)

  (* Ev aa ~: Ev ` X *)
  apply (elim conjE)
  apply (erule disjE)

   (* left step --> contra *)
   apply (elim conjE exE)
   apply (simp add: Abs_trace_inject trace_event_app_in_trace event_in_trace)
   apply (simp add: event_in_trace Abs_trace_inverse)

   (* right step *)
   apply (elim conjE exE)
   apply (insert same_head_tail)
   apply (drule_tac x="a" in spec)
   apply (drule_tac x="aa" in spec)
   apply (drule_tac x="t" in spec)
   apply (drule_tac x="t'" in spec)

   apply (simp add: Rep_trace_inject image_iff)
   apply (rule_tac x="s" in exI)
   apply (simp)
done

(**************************************************
             [Tick]t |[X]|set []t
 **************************************************)

lemma par_trace_Tick_nil:
 "ALL X s. [Tick]t |[X]|set []t = {}"
apply (intro allI impI)
apply (simp add: par_rep_def)
apply (simp add: event_in_trace nil_in_trace Abs_trace_inverse)
apply (simp add: par_list_def)
done

(*** sym ***)

lemma par_trace_nil_Tick:
 "ALL X s. []t |[X]|set [Tick]t = {}"
apply (intro allI impI)
apply (simp add: par_rep_def)
apply (simp add: event_in_trace nil_in_trace Abs_trace_inverse)
apply (simp add: par_list_def)
done

(****************** to add it again ******************)

declare disj_not1 [simp] 

end
